# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

import json
from datetime import datetime, timedelta
from io import BytesIO
from pathlib import Path
from zipfile import ZipFile

import numpy as np
import scipy.stats as stats
import tensorly as tl
from scipy.io import loadmat

try:
    import pandas as pd
    import requests
    from tqdm import tqdm
except ModuleNotFoundError:
    pass

from matcouply.coupled_matrices import CoupledMatrixFactorization

DATASET_PARENT = Path(__file__).parent / "datasets"
DOWNLOADED_PARENT = DATASET_PARENT / "downloads"


def get_simple_simulated_data(noise_level=0.2, random_state=1):
    r"""Generate a simple simulated dataset with shifting unimodal :math:`\mathbf{B}^{(i)}` matrices.

    The entries in :math:`\mathbf{A}` (or :math:`\mathbf{D}^{(i)}`-matrices) are uniformly distributed
    between 0.1 and 1.1. This is done to ensure that there is signal from all components in all matrices.

    The component vectors in the :math:`\mathbf{B}^{(i)}` matrices are Gaussian probability density
    functions that shift one entry for each matrix. This means that they are non-negative, unimodal
    and satisfy the PARAFAC2 constraint.

    The entries in :math:`\mathbf{C}` are truncated normal distributed, and are therefore sparse.

    The dataset is generated by constructing the matrices represented by the decomposition and adding
    noise according to

    .. math::

        \mathbf{M}_\text{noisy}^{(i)} =
        \mathbf{M}^{(i)} + \eta \frac{\|\mathbf{\mathcal{M}}\|}{\|\mathbf{\mathcal{N}}\|} \mathbf{N}^{(i)},


    where :math:`\eta` is the noise level, :math:`\mathbf{M}^{(i)}` is the :math:`i`-th matrix represented
    by the simulated factorization, :math:`\mathbf{\mathcal{M}}` is the tensor obtained by stacking all the
    :math:`\mathbf{M}^{(i)}`-matrices, :math:`n^{(i)}_{jk} \sim \mathcal{N}(0, 1)` and :math:`\mathbf{N}^{(i)}`
    and :math:`\mathbf{\mathcal{N}}` and is the matrix and tensor with elements given by :math:`n^{(i)}_{jk}`, respectively.

    Parameters
    ----------
    noise_level : float
        Strength of noise added to the matrices
    random_state : None, int or valid tensorly random state

    Returns
    -------
    list of matrices
        The noisy matrices
    CoupledMatrixFactorization
        The factorization that underlies the simulated data matrices
    """
    rank = 3
    I, J, K = 15, 50, 20

    rng = tl.check_random_state(random_state)

    # Uniformly distributed A
    A = rng.uniform(size=(I, rank)) + 0.1  # Add 0.1 to ensure that there is signal for all components for all slices
    A = tl.tensor(tl.tensor(A))

    # Generate B-matrix as shifting normal distributions
    t = np.linspace(-10, 10, J)
    B_blueprint = np.stack([stats.norm.pdf(t, loc=-5), stats.norm.pdf(t, loc=0), stats.norm.pdf(t, loc=2)], axis=-1)
    B_is = [np.roll(B_blueprint, i, axis=0) for i in range(I)]  # Cyclically permute to get changing B_i matrices
    B_is = [tl.tensor(B_i) for B_i in B_is]

    # Truncated normal distributed C
    C = tl.tensor(rng.standard_normal(size=(K, rank)))
    C[C < 0] = 0
    C = tl.tensor(C)

    # Construct decomposition and data matrix
    cmf = CoupledMatrixFactorization((None, (A, B_is, C)))
    matrices = cmf.to_matrices()

    # Add noise
    noise = [tl.tensor(rng.standard_normal(size=M.shape)) for M in matrices]
    scale_factor = tl.norm(tl.stack(matrices)) / tl.norm(tl.stack(noise))
    matrices = [M + noise_level * scale_factor * N for M, N in zip(matrices, noise)]
    return matrices, cmf


def get_bike_data():
    r"""Get bike sharing data from three major Norwegian cities

    This dataset contains three matrices with bike sharing data from Oslo, Bergen and Trondheim,
    :math:`\mathbf{X}^{(\text{Oslo})}, \mathbf{X}^{(\text{Bergen})}` and :math:`\mathbf{X}^{(\text{Trondheim})}`.
    Each row of these data matrices represent a station, and each column of the data matrices
    represent an hour in 2021. The matrix element :math:`x^{(\text{Oslo})}_{jk}` is the number of trips
    that ended in station :math:`j` in Oslo during hour :math:`k`.

    The data was obtained using the open API of

     * Oslo Bysykkel: https://oslobysykkel.no/en/open-data
     * Bergen Bysykkel: https://bergenbysykkel.no/en/open-data
     * Trondheim Bysykkel: https://trondheimbysykkel.no/en/open-data

    on the 23rd of November 2021.

    The dataset is cleaned so it only contains for the dates in 2021 when bike sharing was open in all three
    cities (2021-04-07 - 2021-11-23).

    Returns
    -------
    dict
        Dictionary mapping the city name with a data frame that contain bike sharing data from that city.
        There is also an additional ``"station_metadata"``-key, which maps to a data frame with additional
        station metadata. This metadata is useful for interpreting the extracted components.

    .. note::

        The original bike sharing data is released under a NLOD lisence (https://data.norge.no/nlod/en/2.0/).
    """
    with ZipFile(DATASET_PARENT / "bike.zip") as data_zip:
        with data_zip.open("bike.json", "r") as f:
            bike_data = json.load(f)

    datasets = {key: pd.DataFrame(value) for key, value in bike_data["dataset"].items()}
    time = [datetime(2021, 1, 1) + timedelta(hours=int(h)) for h in datasets["oslo"].columns]
    time = pd.to_datetime(time).tz_localize("UTC").tz_convert("CET")

    out = {}
    for city in ["oslo", "trondheim", "bergen"]:
        df = datasets[city]
        df.columns = time
        df.columns.name = "Time of arrival"
        df.index.name = "Arrival station ID"
        df.index = df.index.astype(int)
        out[city] = df

    out["station_metadata"] = datasets["station_metadata"]
    out["station_metadata"].index.name = "Arrival station ID"
    out["station_metadata"].index = out["station_metadata"].index.astype(int)

    return out


def get_semiconductor_etch_raw_data(download_data=True, save_data=True):
    """Load semiconductor etch data from :cite:p:`wise1999comparison`.

    If the dataset is already downloaded on your computer, then the local files will be
    loaded. Otherwise, they will be downloaded. By default, the files are downloaded from
    https://eigenvector.com/data/Etch.

    Parameters
    ----------
    download_data : bool
        If ``False``, then an error will be raised if the data is not
        already downloaded.
    save_data : bool
        if ``True``, then the data will be stored locally to avoid having to download
        multiple times.

    Returns
    -------
    dict
        Dictionary where the keys are the dataset names and the values are the contents
        of the MATLAB files.
    """
    data_urls = {
        "MACHINE_Data.mat": "http://eigenvector.com/data/Etch/MACHINE_Data.mat",
        "OES_DATA.mat": "http://eigenvector.com/data/Etch/OES_DATA.mat",
        "RFM_DATA.mat": "http://eigenvector.com/data/Etch/RFM_DATA.mat",
    }
    data_raw_mat = {}

    print("Loading semiconductor etch data from Wise et al. (1999) - J. Chemom. 13(3‐4), pp.379-396.")
    print("The data is available at: http://eigenvector.com/data/Etch/")
    for file, url in tqdm(data_urls.items()):
        file_path = DOWNLOADED_PARENT / file
        if file_path.exists():
            data_raw_mat[file] = loadmat(file_path)
        elif download_data:
            request = requests.get(url)
            if request.status_code != 200:
                raise RuntimeError(f"Cannot download file {url} - Response: {request.status_code} {request.reason}")

            if save_data:
                DOWNLOADED_PARENT.mkdir(exist_ok=True, parents=True)
                with open(file_path, "wb") as f:
                    f.write(request.content)

            data_raw_mat[file] = loadmat(BytesIO(request.content))
        else:
            raise RuntimeError("The semiconductor etch data is not yet downloaded, and ``download_data=False``.")
    return data_raw_mat


def get_semiconductor_etch_machine_data(download_data=True, save_data=True):
    """Load machine measurements from the semiconductor etch dataset from :cite:p:`wise1999comparison`.

    This function will load the semiconductor etch machine data and prepare it for analysis.

    If the dataset is already downloaded on your computer, then the local files will be
    loaded. Otherwise, they will be downloaded. By default, the files are downloaded from
    https://eigenvector.com/data/Etch.

    Parameters
    ----------
    download_data : bool
        If ``False``, then an error will be raised if the data is not
        already downloaded.
    save_data : bool
        if ``True``, then the data will be stored locally to avoid having to download
        multiple times.

    Returns
    -------
    dict
        Dictionary where the keys are the dataset names and the values are the contents
        of the MATLAB files.
    """
    # Get raw MATLAB data and parse into Python dict
    data = get_semiconductor_etch_raw_data(download_data=download_data, save_data=save_data)["MACHINE_Data.mat"][
        "LAMDATA"
    ]
    data = {key: data[key].squeeze().item().squeeze() for key in data.dtype.fields}

    # Format data as xarray dataset
    varnames = data["variables"][2:]

    # Get the training data
    train_names = [name.split(".")[0][1:] for name in data["calib_names"]]
    train_data = {
        name: pd.DataFrame(Xi[:-1, 2:], columns=varnames)  # Slice away last row since it "belongs" to the next sample
        for name, Xi in zip(train_names, data["calibration"])
    }
    train_metadata = {}
    for i, name in enumerate(list(train_data)):
        train_data[name].index.name = "Time point"
        train_data[name].columns.name = "Measurement"

        metadata = pd.DataFrame(data["calibration"][i][:-1, [0, 1]], columns=["Time", "Step number"])
        metadata["Experiment"] = int(name[:2])
        metadata["Sample"] = int(name[2:])
        metadata.index.name = "Time point"
        metadata.columns.name = "Metadata"
        train_metadata[name] = metadata

    # Get the testing data
    test_names = [name.split(".")[0][1:] for name in data["test_names"]]
    test_data = {
        name: pd.DataFrame(Xi[:-1, 2:], columns=varnames)  # Slice away last row since it "belongs" to the next sample
        for name, Xi in zip(test_names, data["test"])
    }
    test_metadata = {}
    for i, name in enumerate(list(test_data)):
        test_data[name].index.name = "Time point"
        test_data[name].columns.name = "Measurement"

        metadata = pd.DataFrame(data["test"][i][:-1, [0, 1]], columns=["Time", "Step number"])
        metadata["Experiment"] = int(name[:2])
        metadata["Sample"] = int(name[2:])
        metadata["Fault name"] = data["fault_names"][i]
        metadata.index.name = "Time point"
        metadata.columns.name = "Metadata"
        test_metadata[name] = metadata

    return train_data, train_metadata, test_data, test_metadata



# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

import pytest
from .utils import random_length


@pytest.fixture
def seed(pytestconfig):
    try:
        return pytestconfig.getoption("randomly_seed")
    except ValueError:
        return 1


@pytest.fixture
def rng(seed):
    import tensorly as tl

    return tl.check_random_state(seed)


@pytest.fixture
def random_ragged_shapes(rng):
    I = random_length(rng)
    K = random_length(rng)

    shapes = tuple((random_length(rng), K) for i in range(I))
    return shapes


@pytest.fixture
def random_regular_shapes(rng):
    I = random_length(rng)
    J = random_length(rng)
    K = random_length(rng)
    shapes = tuple((J, K) for i in range(I))
    return shapes


@pytest.fixture
def random_ragged_cmf(
    rng, random_ragged_shapes,
):
    from matcouply.random import random_coupled_matrices

    smallest_J = min(shape[0] for shape in random_ragged_shapes)
    rank = rng.randint(1, smallest_J + 1)
    cmf = random_coupled_matrices(random_ragged_shapes, rank, random_state=rng)
    return cmf, random_ragged_shapes, rank


@pytest.fixture
def random_rank5_ragged_cmf(rng):
    from matcouply.random import random_coupled_matrices

    I = random_length(rng)
    K = random_length(rng)
    random_ragged_shapes = tuple((random_length(rng, min=5, mean=7), K) for i in range(I))
    rank = 5
    cmf = random_coupled_matrices(random_ragged_shapes, rank, random_state=rng)
    return cmf, random_ragged_shapes, rank


@pytest.fixture
def random_regular_cmf(
    rng, random_regular_shapes,
):
    from matcouply.random import random_coupled_matrices

    rank = rng.randint(1, random_regular_shapes[0][0] + 1)
    cmf = random_coupled_matrices(random_regular_shapes, rank, random_state=rng)
    return cmf, random_regular_shapes, rank


@pytest.fixture
def random_matrix(rng):
    import tensorly as tl

    return tl.tensor(rng.standard_normal((10, 3)))


@pytest.fixture
def random_matrices(rng):
    import tensorly as tl

    return [tl.tensor(rng.standard_normal((10, 3))) for i in range(5)]



# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.
from unittest.mock import patch

import numpy as np
import pytest
import scipy
import tensorly as tl
from numpy.core.numeric import allclose
from tensorly.testing import assert_array_equal

from matcouply import _utils as utils


def test_is_iterable():
    # Test with some objects that are iterable
    class TestIterable:
        def __iter__(self):
            return self

        def __next__(self):
            return 1

    test_iterables = [
        [1, 2, 3, 4],
        (1, 2, 3, 4),
        "iterablestring",
        {1: "value1", 2: "value2"},
        range(5),
        TestIterable(),
    ]
    for test_iterable in test_iterables:
        assert utils.is_iterable(test_iterable)

    # Test with some objects that arent't iterable
    def test_function(x):
        return x

    class TestNotIterable:
        pass

    test_not_iterables = [1, 3.14, test_function, TestNotIterable()]
    for test_not_iterable in test_not_iterables:
        assert not utils.is_iterable(test_not_iterable)


@pytest.mark.parametrize("svd", ["truncated_svd"])
def test_get_svd(rng, svd):
    X = rng.standard_normal(size=(10, 20))
    U1, s1, Vh1 = scipy.linalg.svd(X)
    svd_fun = utils.get_svd(svd)
    U2, s2, Vh2 = svd_fun(tl.tensor(X))
    U2, s2, Vh2 = tl.to_numpy(U2), tl.to_numpy(s2), tl.to_numpy(Vh2)

    # Check singular values are the same
    assert allclose(s1, s2)

    # Check that first 10 (rank) singular vectors are equal or flipped
    U1TU2 = U1.T @ U2
    Vh1Vh2T = Vh1 @ Vh2.T
    Vh1Vh2T = Vh1Vh2T[:10, :10]
    assert allclose(U1TU2, Vh1Vh2T, atol=1e-6)  # low tolerance due to roundoff errors
    assert allclose(U1TU2 * Vh1Vh2T, np.eye(U1TU2.shape[0]))


def test_get_svd_works_with_old_tensorly():
    svds = {"TEST": "SVD"}
    with patch("matcouply._utils.tl.SVD_FUNS", svds) as mock:
        svd = utils.get_svd("TEST")

    assert svd == svds["TEST"]


def test_get_svd_fails_with_invalid_svd_name():
    with pytest.raises(ValueError):
        utils.get_svd("THIS_IS_NOT_A_VALID_SVD")


def test_get_shapes(rng, random_ragged_cmf):
    # Small manual test
    matrices = [tl.zeros((1, 2)), tl.zeros((3, 4)), tl.zeros((5, 6))]
    matrix_shapes = utils.get_shapes(matrices)
    assert matrix_shapes[0] == (1, 2)
    assert matrix_shapes[1] == (3, 4)
    assert matrix_shapes[2] == (5, 6)
    assert len(matrix_shapes) == 3

    # Test on random ragged cmf
    cmf, shapes, rank = random_ragged_cmf
    matrix_shapes = utils.get_shapes(cmf.to_matrices())
    for matrix_shape, shape in zip(matrix_shapes, shapes):
        assert matrix_shape == shape


def test_get_padded_tensor_shape(rng, random_ragged_cmf):
    cmf, shapes, rank = random_ragged_cmf

    I = len(shapes)
    J = max([shape[0] for shape in shapes])
    K = shapes[0][1]

    assert (I, J, K) == utils.get_padded_tensor_shape(cmf.to_matrices())

    matrices_different_columns = [
        tl.tensor(rng.standard_normal(size=(3, 4))),
        tl.tensor(rng.standard_normal(size=(5, 6))),
        tl.tensor(rng.standard_normal(size=(5, 6))),
    ]
    with pytest.raises(ValueError):
        utils.get_padded_tensor_shape(matrices_different_columns)


def test_create_padded_tensor(rng, random_ragged_cmf):
    cmf, shapes, rank = random_ragged_cmf
    matrices = cmf.to_matrices()
    padded_tensor = utils.create_padded_tensor(matrices)

    I = len(shapes)
    J = max([shape[0] for shape in shapes])
    K = shapes[0][1]

    assert (I, J, K) == tl.shape(padded_tensor)

    for i, (matrix, shape) in enumerate(zip(matrices, shapes)):
        assert_array_equal(padded_tensor[i, : shape[0], :], matrix)
        assert_array_equal(padded_tensor[i, shape[0] :, :], 0)



# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

from unittest.mock import MagicMock, patch

import numpy as np
import pytest
import tensorly as tl
from tensorly.testing import assert_array_equal

from matcouply import data
from matcouply.coupled_matrices import CoupledMatrixFactorization
from matcouply.decomposition import _cmf_reconstruction_error

from .utils import RTOL_SCALE


def test_get_simple_simulated_data():
    simulated_data, cmf = data.get_simple_simulated_data()
    assert isinstance(simulated_data, list)
    assert isinstance(cmf, CoupledMatrixFactorization)

    # Check that same random state gets same data
    simulated_data1, cmf1 = data.get_simple_simulated_data(random_state=1)
    simulated_data2, cmf2 = data.get_simple_simulated_data(random_state=1)

    for X1, X2 in zip(simulated_data1, simulated_data2):
        assert_array_equal(X1, X2)

    assert_array_equal(cmf1[1][0], cmf2[1][0])  # check A
    assert_array_equal(cmf1[1][2], cmf2[1][2])  # check C

    for B_i1, B_i2 in zip(cmf1[1][1], cmf2[1][1]):
        assert_array_equal(B_i1, B_i2)

    # Check that different random state gets different data
    simulated_data3, cmf3 = data.get_simple_simulated_data(random_state=2)

    assert not np.all(tl.to_numpy(cmf1[1][0]) == tl.to_numpy(cmf3[1][0]))
    assert not np.all(tl.to_numpy(cmf1[1][2]) == tl.to_numpy(cmf3[1][2]))

    # Check that noise level is correct
    simulated_data_noise_02, cmf_noise_02 = data.get_simple_simulated_data(noise_level=0.2, random_state=2)
    data_strength = tl.norm(cmf_noise_02.to_tensor())
    error = _cmf_reconstruction_error(simulated_data_noise_02, cmf_noise_02) / data_strength
    assert error == pytest.approx(0.2, rel=1e-6*RTOL_SCALE)


def test_get_bike_data():
    bike_data = data.get_bike_data()

    # Check that data has correct keys
    cities = ["oslo", "trondheim", "bergen"]

    # Check that data has correct columns
    stations = set(bike_data["station_metadata"].index)
    for city in cities:
        assert city in bike_data
        assert bike_data[city].shape[1] == bike_data["oslo"].shape[1]
        assert all(bike_data[city].columns == bike_data["oslo"].columns)

        for station in bike_data[city].index:
            assert station in stations


def test_get_etch_data_prints_citation(capfd):
    data.get_semiconductor_etch_raw_data()
    assert "Wise et al. (1999) - J. Chemom. 13(3‐4)" in capfd.readouterr()[0]

    data.get_semiconductor_etch_machine_data()
    assert "Wise et al. (1999) - J. Chemom. 13(3‐4)" in capfd.readouterr()[0]


@pytest.fixture
def patch_dataset_parent(tmp_path):
    old_path = data.DATASET_PARENT
    old_download_parent = data.DOWNLOADED_PARENT
    data.DATASET_PARENT = tmp_path
    data.DOWNLOADED_PARENT = tmp_path / "downloads"
    yield old_path
    data.DATASET_PARENT = old_path
    data.DOWNLOADED_PARENT = old_download_parent


class MockSuccessfullRequest(MagicMock):
    status_code = 200
    content = b""


class MockUnsuccessfullRequest(MagicMock):
    status_code = 404
    content = b""


@patch("matcouply.data.requests.get", return_value=MockSuccessfullRequest())
@patch("matcouply.data.loadmat")
@patch("matcouply.data.BytesIO")
def test_get_etch_raw_data_downloads_correctly(bytesio_mock, loadmat_mock, get_mock, patch_dataset_parent):
    with pytest.raises(RuntimeError):
        data.get_semiconductor_etch_raw_data(download_data=False)

    # Check that it works once when we download but don't save
    data.get_semiconductor_etch_raw_data(save_data=False)

    # Check that it wasn't saved when save_data=False
    with pytest.raises(RuntimeError):
        data.get_semiconductor_etch_raw_data(download_data=False)

    # Check that it raises error with unsuccessful download
    get_mock.return_value = MockUnsuccessfullRequest()
    with pytest.raises(RuntimeError):
        data.get_semiconductor_etch_raw_data()

    # Check that it works once the data is downloaded
    get_mock.return_value = MockSuccessfullRequest()
    data.get_semiconductor_etch_raw_data()
    data.get_semiconductor_etch_raw_data(download_data=False)


def test_get_semiconductor_etch_raw_data():
    raw_data = data.get_semiconductor_etch_raw_data()
    for file in ["MACHINE_Data.mat", "OES_DATA.mat", "RFM_DATA.mat"]:
        assert file in raw_data


def test_get_semiconductor_machine_data():
    train_data, train_metadata, test_data, test_metadata = data.get_semiconductor_etch_machine_data()

    train_sample_names = set(train_data)
    test_sample_names = set(test_data)
    train_metadata_names = set(train_metadata)
    test_metadata_names = set(test_metadata)

    assert len(train_sample_names.intersection(test_sample_names)) == 0
    assert len(train_sample_names.intersection(train_metadata_names)) == len(train_sample_names)
    assert len(test_sample_names.intersection(test_metadata_names)) == len(test_sample_names)

    # Check that data has correct columns
    one_train_name = next(iter(train_sample_names))
    for name in train_data:
        assert train_data[name].shape[1] == train_data[one_train_name].shape[1]
        assert train_metadata[name].shape[0] == train_data[name].shape[0]
        assert all(train_data[name].columns == train_data[one_train_name].columns)

    one_test_name = next(iter(test_sample_names))
    for name in test_data:
        assert test_data[name].shape[1] == test_data[one_test_name].shape[1]
        assert test_metadata[name].shape[0] == test_data[name].shape[0]
        assert all(test_data[name].columns == test_data[one_test_name].columns)



# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

from copy import copy
from typing import NamedTuple, Optional

import numpy as np
import tensorly as tl
import tensorly.decomposition

from . import penalties
from ._utils import create_padded_tensor, get_shapes, get_svd, is_iterable
from .coupled_matrices import CoupledMatrixFactorization, cmf_to_matrices

__all__ = ["compute_feasibility_gaps", "ADMMVars", "DiagnosticMetrics", "cmf_aoadmm", "parafac2_aoadmm"]


def _update_imputed(tensor_slices, mask, decomposition, method):
    """
    Update missing values of tensor slices according to method.

    Parameters
    ----------
    tensor_slices : Iterable of ndarray
    mask : ndarray
        An array with the same shape as the tensor. It should be 0 where there are
        missing values and 1 everywhere else.
    decomposition : Parafac2Tensor, optional
    method : string
        One of 'mode-3' or 'factors'. 'mode-3' updates imputed values according to
        mean of each mode-3 slice. If 'factors' is chosen, set missing entries
        according to reconstructed tensor given from 'decomposition'.
        'mode-3' is used (by default) for initializing missing entries while
        'factors' is used for updating imputations during optimization. If an
        initial decomposition is specified, 'factors' is used at initialization.

    Returns
    -------
    tensor_slices : Iterable of ndarray
    """
    if method == "mode-3":
        for slice_no, (slice, slice_mask) in enumerate(zip(tensor_slices, mask)):
            slice_mean = tl.sum(slice * slice_mask) / tl.sum(slice_mask)
            tensor_slices[slice_no] = tl.where(slice_mask == 0, slice_mean, tensor_slices[slice_no])

    else:
        reconstructed_slices = cmf_to_matrices(decomposition)
        tensor_slices = list(tensor_slices)

        for slice_no, (slice, rec_slice, slice_mask) in enumerate(zip(tensor_slices, reconstructed_slices, mask)):
            tensor_slices[slice_no] = tl.where(slice_mask == 0, rec_slice, slice)

    return tensor_slices


def initialize_cmf(matrices, rank, init, svd_fun, random_state=None, init_params=None):
    random_state = tl.check_random_state(random_state)

    # Start by checking if init is a valid factorization. If so, use it
    if isinstance(init, (tuple, list, CoupledMatrixFactorization)):
        weights, (A, B_is, C) = init
        if weights is not None:
            scaled_A = weights * A
            return CoupledMatrixFactorization((None, (scaled_A, B_is, C)))
        else:
            return CoupledMatrixFactorization(init)

    # Random uniform initialisation
    if init == "random":
        I = len(matrices)
        K = tl.shape(matrices[0])[1]

        A = tl.tensor(random_state.uniform(size=(I, rank)))
        C = tl.tensor(random_state.uniform(size=(K, rank)))
        B_is = [tl.tensor(random_state.uniform(size=(tl.shape(matrix)[0], rank))) for matrix in matrices]

        return CoupledMatrixFactorization((None, [A, B_is, C]))

    # SVD and thresholded SVD initialisation
    if init == "svd" or init == "threshold_svd":
        I = len(matrices)
        A = tl.ones((I, rank))
        B_is = [svd_fun(matrix, n_eigenvecs=rank)[0] for matrix in matrices]
        C = tl.transpose(svd_fun(tl.concatenate(matrices, 0), n_eigenvecs=rank)[2])
        if init == "svd":
            return CoupledMatrixFactorization((None, [A, B_is, C]))

        A = tl.ones((I, rank))
        B_is = [tl.clip(B_i, 0, float("inf")) for B_i in B_is]
        C = tl.clip(C, 0, float("inf"))
        return CoupledMatrixFactorization((None, [A, B_is, C]))

    # PARAFAC and PARAFAC2 initialisation:
    if init_params is None:
        init_params = {}
    if "n_iter_max" not in init_params:
        init_params["n_iter_max"] = 50

    if init == "parafac2_als":
        pf2 = tl.decomposition.parafac2(matrices, rank, **init_params, random_state=random_state)
        return CoupledMatrixFactorization.from_Parafac2Tensor(pf2)

    # PARAFAC init: Work with padded tensor
    tensor = create_padded_tensor(matrices)
    shapes = get_shapes(matrices)
    if init == "cp_als" or init == "parafac_als":
        cp = tl.decomposition.parafac(tensor, rank, **init_params, random_state=random_state)
        return CoupledMatrixFactorization.from_CPTensor(cp, shapes=shapes)
    if init == "cp_hals" or init == "parafac_hals":
        cp = tl.decomposition.non_negative_parafac_hals(tensor, rank, **init_params, random_state=random_state)
        return CoupledMatrixFactorization.from_CPTensor(cp, shapes=shapes)

    raise ValueError('Initialization method "{}" not recognized'.format(init))


def initialize_aux(matrices, rank, reg, random_state):
    A_aux_list = [A_reg.init_aux(matrices, rank, 0, random_state=random_state) for A_reg in reg[0]]
    B_aux_list = [B_reg.init_aux(matrices, rank, 1, random_state=random_state) for B_reg in reg[1]]
    C_aux_list = [C_reg.init_aux(matrices, rank, 2, random_state=random_state) for C_reg in reg[2]]
    return A_aux_list, B_aux_list, C_aux_list


def initialize_dual(matrices, rank, reg, random_state):
    A_dual_list = [A_reg.init_dual(matrices, rank, 0, random_state=random_state) for A_reg in reg[0]]
    B_dual_list = [B_reg.init_dual(matrices, rank, 1, random_state=random_state) for B_reg in reg[1]]
    C_dual_list = [C_reg.init_dual(matrices, rank, 2, random_state=random_state) for C_reg in reg[2]]
    return A_dual_list, B_dual_list, C_dual_list


def _check_inner_convergence(factor_matrix, old_factor_matrix, cmf, reg_list, aux_list, mode, inner_tol):
    if not inner_tol or inner_tol < 0:
        return False

    if mode == 1:
        norm = _root_sum_squared_list(factor_matrix)
        change = _root_sum_squared_list([B_i - prev_B_i for B_i, prev_B_i in zip(factor_matrix, old_factor_matrix)])
    else:
        norm = tl.norm(factor_matrix)
        change = tl.norm(factor_matrix - old_factor_matrix)

    if change > inner_tol * norm:
        return False

    if len(reg_list) == 0:
        return True

    # Create regs-list and auxes-list so only feasibility gaps for current mode is computed
    regs = [[], [], []]
    regs[mode] = reg_list
    auxes = [[], [], []]
    auxes[mode] = aux_list
    gaps = compute_feasibility_gaps(cmf, regs, *auxes)[mode]

    return max(gaps) < inner_tol


# TODO (Improvement): Add option to scale the l2_penalty based on size (mostly relevant for B)
def admm_update_A(
    matrices,
    reg,
    cmf,
    A_aux_list,
    A_dual_list,
    l2_penalty,
    inner_n_iter_max,
    inner_tol,
    feasibility_penalty_scale,
    constant_feasibility_penalty,
    svd_fun,
):
    weights, (A, B_is, C) = cmf

    # Compute cross products (B_i^T B_i * C^T C) and RHS
    cross_products = []
    rhses = []
    CtC = tl.dot(tl.transpose(C), C)

    for matrix, B_i in zip(matrices, B_is):
        # Multiply three matrices (B_i X_i C) in the most efficient order:
        # (B_i X_i) C has complexity O((K J_i R) + (K R^2))
        # B_i (X_i C) has complexity O((J_i R^2) + (K J_i R))
        # By checking I and J_i, we can compute the product in the most efficient order
        J_i = tl.shape(B_i)[0]
        I = tl.shape(C)[0]
        if J_i > I:
            tmp = tl.dot(tl.transpose(B_i), matrix)
            BtXC_i = tl.dot(tmp, C)
        else:
            tmp = tl.dot(matrix, C)
            BtXC_i = tl.dot(tl.transpose(B_i), tmp)

        # Store cross products for efficient SSE computation
        cross_products.append(tl.dot(tl.transpose(B_i), B_i) * CtC)

        # Construct RHS
        rhses.append(tl.diag(BtXC_i))

    # Multiply with 0.5 since this function minimizes 0.5||Ax - b||^2
    # while in the PARAFAC2 AO-ADMM paper ||Ax - b||^2 is minimzed
    feasibility_penalties = [0.5 * tl.trace(lhs) * feasibility_penalty_scale for lhs in cross_products]
    if constant_feasibility_penalty:
        max_feasibility_penalty = max(feasibility_penalties)
        feasibility_penalties = [max_feasibility_penalty for _ in feasibility_penalties]

    # Construct LHS and compute its SVD for efficiently solving for it
    lhses = [
        lhs + tl.eye(tl.shape(A)[1]) * (feasibility_penalty * len(reg) + l2_penalty)
        for lhs, feasibility_penalty in zip(cross_products, feasibility_penalties)
    ]
    svds = [svd_fun(lhs) for lhs in lhses]

    # Inner ADMM loop for updating A
    old_A = tl.copy(A)
    for inner_it in range(inner_n_iter_max):
        old_A, A = A, old_A

        shifted_A_auxes = [
            single_reg.subtract_from_aux(A_aux, A_dual)
            for single_reg, A_aux, A_dual in zip(reg, A_aux_list, A_dual_list)
        ]

        for i, matrix in enumerate(matrices):
            U, s, Uh = svds[i]

            sum_shifted_aux_A_row = 0
            for shifted_A_aux in shifted_A_auxes:
                sum_shifted_aux_A_row += shifted_A_aux[i]

            tl.index_update(
                A,
                tl.index[i, :],
                tl.dot(tl.dot(feasibility_penalties[i] * sum_shifted_aux_A_row + rhses[i], U / s), Uh),
            )

        for reg_num, single_reg in enumerate(reg):
            A_aux = A_aux_list[reg_num]
            A_dual = A_dual_list[reg_num]
            shifted_A = A + A_dual

            if constant_feasibility_penalty:
                A_aux_list[reg_num] = single_reg.factor_matrix_update(shifted_A, max_feasibility_penalty, A_aux)

            else:
                for i, feasibility_penalty in enumerate(feasibility_penalties):
                    tl.index_update(
                        A_aux,
                        tl.index[i, :],
                        single_reg.factor_matrix_row_update(shifted_A[i], feasibility_penalty, A_aux[i]),
                    )
            shifted_A_auxes[reg_num] = single_reg.subtract_from_aux(A_aux_list[reg_num], A_dual)
            A_dual_list[reg_num] = A - shifted_A_auxes[reg_num]  # A - (A_aux - A_dual) = A - A_aux + A_dual

        cmf = weights, (A, B_is, C)
        if _check_inner_convergence(A, old_A, cmf, reg, A_aux_list, mode=0, inner_tol=inner_tol):
            break

    return (None, [A, B_is, C]), A_aux_list, A_dual_list, (rhses, cross_products)


def admm_update_B(
    matrices,
    reg,
    cmf,
    B_is_aux_list,
    B_is_dual_list,
    l2_penalty,
    inner_n_iter_max,
    inner_tol,
    feasibility_penalty_scale,
    constant_feasibility_penalty,
    svd_fun,
):
    weights, (A, B_is, C) = cmf

    # Compute lhs and rhs
    lhses = []
    rhses = []
    CtC = tl.dot(tl.transpose(C), C)
    for matrix, a_row in zip(matrices, A):
        rhses.append(tl.dot(matrix, C * a_row))
        lhses.append(tl.transpose(tl.transpose(CtC * a_row) * a_row))

    # Multiply with 0.5 since this function minimizes 0.5||Ax - b||^2
    # while in the PARAFAC2 AO-ADMM paper ||Ax - b||^2 is minimzed
    feasibility_penalties = [0.5 * tl.trace(lhs) * feasibility_penalty_scale for lhs in lhses]
    if constant_feasibility_penalty:
        max_feasibility_penalty = max(feasibility_penalties)
        feasibility_penalties = [max_feasibility_penalty for _ in feasibility_penalties]

    lhses = [
        lhs + tl.eye(tl.shape(A)[1]) * (feasibility_penalty * len(reg) + l2_penalty)
        for lhs, feasibility_penalty in zip(lhses, feasibility_penalties)
    ]
    svds = [svd_fun(lhs) for lhs in lhses]

    old_B_is = [tl.copy(B_i) for B_i in B_is]
    for inner_it in range(inner_n_iter_max):
        old_B_is, B_is = B_is, old_B_is
        shifted_auxes_B_is = [
            single_reg.subtract_from_auxes(B_is_aux, B_is_dual)
            for single_reg, B_is_aux, B_is_dual in zip(reg, B_is_aux_list, B_is_dual_list)
        ]

        for i, matrix in enumerate(matrices):
            U, s, Uh = svds[i]

            sum_shifted_aux_B_i = 0
            for shifted_aux_B_is in shifted_auxes_B_is:
                sum_shifted_aux_B_i += shifted_aux_B_is[i]  # B_is_aux[i] - B_is_dual[i]

            B_is[i] = tl.dot(tl.dot(feasibility_penalties[i] * sum_shifted_aux_B_i + rhses[i], U / s), Uh)

        for reg_num, single_reg in enumerate(reg):
            B_is_aux = B_is_aux_list[reg_num]
            B_is_dual = B_is_dual_list[reg_num]
            shifted_B_is = [B_i + B_i_dual for B_i, B_i_dual in zip(B_is, B_is_dual)]

            B_is_aux_list[reg_num] = single_reg.factor_matrices_update(shifted_B_is, feasibility_penalties, B_is_aux)

            shifted_auxes_B_is[reg_num] = single_reg.subtract_from_auxes(B_is_aux_list[reg_num], B_is_dual)
            B_is_dual_list[reg_num] = [
                B_i - shifted_aux_B_i for B_i, shifted_aux_B_i in zip(B_is, shifted_auxes_B_is[reg_num])
            ]

        cmf = weights, (A, B_is, C)

        if _check_inner_convergence(B_is, old_B_is, cmf, reg, B_is_aux_list, mode=1, inner_tol=inner_tol):
            break

    return (None, [A, B_is, C]), B_is_aux_list, B_is_dual_list


def admm_update_C(
    matrices,
    reg,
    cmf,
    C_aux_list,
    C_dual_list,
    l2_penalty,
    inner_n_iter_max,
    inner_tol,
    feasibility_penalty_scale,
    svd_fun,
):
    weights, (A, B_is, C) = cmf

    # Compute lhs and rhs
    lhs = 0
    rhs = 0
    for matrix, B_i, A_i in zip(matrices, B_is, A):
        B_iA_i = B_i * A_i
        lhs += tl.dot(tl.transpose(B_iA_i), B_iA_i)
        rhs += tl.dot(tl.transpose(matrix), B_iA_i)

    # Multiply with 0.5 since this function minimizes 0.5||Ax - b||^2
    # while in the PARAFAC2 AO-ADMM paper ||Ax - b||^2 is minimzed
    feasibility_penalty = 0.5 * tl.trace(lhs) * feasibility_penalty_scale
    lhs += tl.eye(tl.shape(C)[1]) * (feasibility_penalty * len(reg) + l2_penalty)
    U, s, Uh = svd_fun(lhs)

    old_C = tl.copy(C)
    # ADMM iterations
    for inner_it in range(inner_n_iter_max):
        old_C, C = C, old_C

        sum_shifted_aux_C = 0
        for single_reg, C_aux, C_dual in zip(reg, C_aux_list, C_dual_list):
            sum_shifted_aux_C += single_reg.subtract_from_aux(C_aux, C_dual)
        C = tl.dot(tl.dot(sum_shifted_aux_C * feasibility_penalty + rhs, U / s), Uh)

        for reg_num, single_reg in enumerate(reg):
            C_aux = C_aux_list[reg_num]
            C_dual = C_dual_list[reg_num]

            C_aux_list[reg_num] = single_reg.factor_matrix_update(C + C_dual, feasibility_penalty, C_aux)
            C_dual_list[reg_num] = C - single_reg.subtract_from_aux(C_aux_list[reg_num], C_dual)

        cmf = weights, (A, B_is, C)
        if _check_inner_convergence(C, old_C, cmf, reg, C_aux_list, mode=2, inner_tol=inner_tol):
            break

    return (None, [A, B_is, C]), C_aux_list, C_dual_list


def _root_sum_squared_list(x_list, mask=None):
    if mask is None:
        return np.sqrt(tl.to_numpy(sum(tl.sum(x**2) for x in x_list)))
    else:
        return np.sqrt(tl.to_numpy(sum(tl.sum(w * x**2) for x, w in zip(x_list, mask))))


def compute_feasibility_gaps(cmf, regs, A_aux_list, B_aux_list, C_aux_list):
    r"""Compute all feasibility gaps.

    The feasibility gaps for AO-ADMM are given by

    .. math::

        \frac{\|\mathbf{x} - \mathbf{z}\|_2}{\|\mathbf{z}\|_2},

    where :math:`\mathbf{x}` is a component vector and :math:`\mathbf{z}` is an auxiliary
    variable that represents a component vector. If a decomposition obtained with AO-ADMM
    is valid, then all feasibility gaps should be small compared to the components. If any
    of the feasibility penalties are large, then the decomposition may not satisfy the
    imposed constraints.

    To compute the feasibility gap for the :math:`\mathbf{A}` and :math:`\mathbf{C}`
    matrices, we use the frobenius norm, and to compute the feasibility gap for the
    :math:`\mathbf{B}^{(i)}`-matrices, we compute :math:`\sqrt{\sum_i \|\mathbf{B}^{(i)} - \mathbf{Z}^{(\mathbf{B}^{(i)})}\|^2}`,
    where :math:`\mathbf{Z}^{(\mathbf{B}^{(i)})}\|^2` is the auxiliary variable for
    :math:`\mathbf{B}^{(i)}`.

    Parameters
    ----------
    cmf: CoupledMatrixFactorization - (weights, factors)
        Coupled matrix factorization represented by weights and factors as described in :doc:`../coupled_matrix_factorization`.

        * weights : 1D array of shape (rank,) or None
            weights of the factors
        * factors : List of factors of the coupled matrix decomposition
            List on the form ``[A, [B_0, B_1, ..., B_i], C]``, where ``A`` represents :math:`\mathbf{A}`,
            ``[B_0, B_1, ..., B_i]`` represents a list of all :math:`\mathbf{B}^{(i)}`-matrices and ``C``
            represents :math:`\mathbf{C}`

    regs : list of list of matcouply.penalties.ADMMPenalty
        The regs list should have three elements, the first being a list of penalties imposed
        on mode 0, the second being a list of penalties imposed on mode 1 and the last being
        a list of penalties imposed on mode 2.
    A_aux_list : list
        A list of all auxiliary variables for the A-matrix
    B_aux_list : list
        A list of all auxiliary variables for the B_is-matrices
    C_aux_list : list
        A list of all auxiliary variables for the C-matrix

    Returns
    -------
    list
        Feasibility gaps for A
    list
        Feasibility gaps for B_is
    list
        Veasibility gaps for C
    """
    weights, (A, B_is, C) = cmf

    A_norm = tl.norm(A)
    B_norm = _root_sum_squared_list(B_is)
    C_norm = tl.norm(C)

    A_gaps = [tl.norm(A_reg.subtract_from_aux(A_aux, A)) / A_norm for A_reg, A_aux in zip(regs[0], A_aux_list)]
    B_gaps = [
        _root_sum_squared_list(B_reg.subtract_from_auxes(B_is_aux, B_is)) / B_norm
        for B_reg, B_is_aux in zip(regs[1], B_aux_list)
    ]
    C_gaps = [tl.norm(C_reg.subtract_from_aux(C_aux, C)) / C_norm for C_reg, C_aux in zip(regs[2], C_aux_list)]

    return A_gaps, B_gaps, C_gaps


def _cmf_reconstruction_error(matrices, cmf, norm_matrices=None, intermediate_A_calculations=None,mask=None):

    if mask is None:
        return _cmf_reconstruction_error_without_mask(matrices,cmf,norm_matrices,intermediate_A_calculations)
    else:
        return _cmf_reconstruction_error_with_mask(matrices,cmf,mask)


def _cmf_reconstruction_error_without_mask(matrices, cmf, norm_matrices=None, intermediate_A_calculations=None):

    if norm_matrices is None:
        norm_X_sq = sum(tl.sum(matrix**2) for matrix in matrices)
    else:
        norm_X_sq = norm_matrices**2

    weights, (A, B_is, C) = cmf
    if weights is not None:
        A = A * weights

    if intermediate_A_calculations is None:
        norm_cmf_sq = 0
        inner_product = 0
        CtC = tl.dot(tl.transpose(C), C)

        for i, B_i in enumerate(B_is):
            B_i = B_i * A[i]
            if tl.shape(B_i)[0] > tl.shape(C)[0]:
                tmp = tl.dot(tl.transpose(B_i), matrices[i])
                inner_product += tl.trace(tl.dot(tmp, C))
            else:
                tmp = tl.dot(matrices[i], C)
                inner_product += tl.trace(tl.dot(tl.transpose(B_i), tmp))

            norm_cmf_sq += tl.sum((B_i.T @ B_i) * CtC)
    else:
        A_rhses, cross_products = intermediate_A_calculations

        inner_product = sum(tl.sum(A_rhs_i * ai) for A_rhs_i, ai in zip(A_rhses, A))
        norm_cmf_sq = sum(tl.sum(tl.diag(ai) @ cross_products[i] @ tl.diag(ai)) for i, ai in enumerate(A))

    # Threshold to handle roundoff errors with very small error
    return np.sqrt(tl.to_numpy(max(0, norm_X_sq - 2 * inner_product + norm_cmf_sq)))

    
def _cmf_reconstruction_error_with_mask(matrices, cmf, mask=None):

    reconstructed_matrices = cmf_to_matrices(cmf)
    total_error = 0
    for i, (slice, slice_mask) in enumerate(zip(matrices, mask)):
        total_error += tl.norm(slice_mask * slice - slice_mask * reconstructed_matrices[i]) ** 2

    return tl.sqrt(total_error)


def _listify(input_value, param_name):
    if hasattr(input_value, "get"):
        return [input_value.get(i, None) for i in range(3)]
    elif not is_iterable(input_value):
        return [input_value] * 3
    else:
        out = list(input_value)
        if not len(out) == 3:
            raise ValueError(
                "All parameters must be a dictionary, non-iterable value or non-dictionary iterable of length 3."
                f" {param_name} is iterable of length {len(out)}."
            )
        return out


def _parse_all_penalties(
    non_negative,
    lower_bound,
    upper_bound,
    l2_norm_bound,
    unimodal,
    parafac2,
    l1_penalty,
    tv_penalty,
    generalized_l2_penalty,
    svd,
    regs,
    dual_init,
    aux_init,
    verbose,
):
    if regs is None:
        regs = [[], [], []]
    elif is_iterable(regs):
        for modereg in regs:
            if not is_iterable(modereg):
                raise TypeError(
                    "regs should contain an iterable of iterables containting "
                    "matcouply.penalties.ADMMMPenalty instances at least one of the"
                    f"elements in regs were not iterable (regs={regs})"
                )
            else:
                for reg in modereg:
                    if not isinstance(reg, penalties.ADMMPenalty):
                        raise TypeError(
                            "regs should contain an iterable of iterables containting "
                            "matcouply.penalties.ADMMMPenalty instances at least one of the"
                            f"elements in regs contained something other than an ADMMPenalty (regs={regs})"
                        )

    regs = [copy(reg_list) for reg_list in regs]  # To avoid side effects where the input lists are modified

    non_negative = _listify(non_negative, "non_negative")
    upper_bound = _listify(upper_bound, "upper_bound")
    lower_bound = _listify(lower_bound, "lower_bound")
    l2_norm_bound = _listify(l2_norm_bound, "l2_norm_bound")
    unimodal = _listify(unimodal, "unimodal")
    if parafac2:
        parafac2 = [False, True, False]
    else:
        parafac2 = [False, False, False]
    l1_penalty = _listify(l1_penalty, "l1_penalty")
    generalized_l2_penalty = _listify(generalized_l2_penalty, "generalized_l2_penalty")
    tv_penalty = _listify(tv_penalty, "tv_penalty")

    for mode in range(3):
        parsed_regs = _parse_mode_penalties(
            non_negative=non_negative[mode],
            lower_bound=lower_bound[mode],
            upper_bound=upper_bound[mode],
            l2_norm_bound=l2_norm_bound[mode],
            unimodal=unimodal[mode],
            parafac2=parafac2[mode],
            l1_penalty=l1_penalty[mode],
            tv_penalty=tv_penalty[mode],
            generalized_l2_penalty=generalized_l2_penalty[mode],
            svd=svd,
            dual_init=dual_init,
            aux_init=aux_init,
        )

        regs[mode] = parsed_regs + regs[mode]

    if verbose:
        print(f"All regularization penalties (including regs list):")
        for mode, reg in enumerate(regs):
            print(f"* Mode {mode}:")
            if len(reg) == 0:
                print("   - (no regularization added)")
            for single_reg in reg:
                print(f"   - {single_reg}")
    return regs


def _parse_mode_penalties(
    non_negative,
    lower_bound,
    upper_bound,
    l2_norm_bound,
    unimodal,
    parafac2,
    l1_penalty,
    tv_penalty,
    generalized_l2_penalty,
    svd,
    dual_init,
    aux_init,
):

    if not l1_penalty:
        l1_penalty = 0

    regs = []

    skip_non_negative = False

    if parafac2:
        regs.append(penalties.Parafac2(svd=svd, aux_init=aux_init, dual_init=dual_init))

    if unimodal:
        regs.append(penalties.Unimodality(non_negativity=non_negative, aux_init=aux_init, dual_init=dual_init))
        skip_non_negative = True

    if (
        generalized_l2_penalty is not None and generalized_l2_penalty is not False
    ):  # generalized_l2_penalty is None or matrix
        regs.append(
            penalties.GeneralizedL2Penalty(generalized_l2_penalty, aux_init=aux_init, dual_init=dual_init, svd=svd)
        )

    if l2_norm_bound:
        regs.append(
            penalties.L2Ball(l2_norm_bound, non_negativity=non_negative, aux_init=aux_init, dual_init=dual_init)
        )
        skip_non_negative = True

    if tv_penalty:
        regs.append(
            penalties.TotalVariationPenalty(tv_penalty, l1_strength=l1_penalty, aux_init=aux_init, dual_init=dual_init)
        )
        l1_penalty = 0  # Already included in the total variation penalty

    if l1_penalty:
        regs.append(
            penalties.L1Penalty(l1_penalty, non_negativity=non_negative, aux_init=aux_init, dual_init=dual_init)
        )
        skip_non_negative = True

    if lower_bound is not None or upper_bound is not None:
        if lower_bound is None:
            lower_bound = -float("inf")
        if non_negative:
            lower_bound = max(lower_bound, 0)
        regs.append(penalties.Box(lower_bound, upper_bound, aux_init=aux_init, dual_init=dual_init))
        skip_non_negative = True

    if non_negative and not skip_non_negative:
        regs.append(penalties.NonNegativity(aux_init=aux_init, dual_init=dual_init))

    return regs


def _compute_l2_penalty(cmf, l2_parameters):
    weights, (A, B_is, C) = cmf
    l2reg = 0
    if l2_parameters[0]:
        l2reg += 0.5 * l2_parameters[0] * tl.sum(A**2)
    if l2_parameters[1]:
        l2reg += 0.5 * l2_parameters[1] * sum(tl.sum(B_i**2) for B_i in B_is)
    if l2_parameters[2]:
        l2reg += 0.5 * l2_parameters[2] * tl.sum(C**2)

    return l2reg


def _check_feasibility(feasibility_gaps, feasibility_tol):
    A_gaps, B_gaps, C_gaps = feasibility_gaps
    max_feasibility_gap = -float("inf")
    if len(A_gaps):
        max_feasibility_gap = max((max(A_gaps), max_feasibility_gap))
    if len(B_gaps):
        max_feasibility_gap = max((max(B_gaps), max_feasibility_gap))
    if len(C_gaps):
        max_feasibility_gap = max((max(C_gaps), max_feasibility_gap))

    return max_feasibility_gap < feasibility_tol


class ADMMVars(NamedTuple):
    auxes: tuple  #: Length three tuple containing a list of auxiliary factor matrices for each mode
    duals: tuple  #: Length three tuple containing a list of dual variables for each mode


class DiagnosticMetrics(NamedTuple):
    rec_errors: list  #: List of length equal to the number of iterations plus one containing the reconstruction errors
    feasibility_gaps: list  #: List of length equal to the number of iterations plus one containing the feasibility gaps
    regularized_loss: list  #: List of length equal to the number of iterations plus one containing the regularized loss
    satisfied_stopping_condition: Optional[
        bool
    ]  #: Boolean specifying whether the stopping conditions were satisfied, None if no tolerance is set
    satisfied_feasibility_condition: Optional[
        bool
    ]  #: Boolean specifying whether the feasibility conditions were satisfied, None if no tolerance is set
    n_iter: int  #: Number of iterations ran
    message: str  #: Convergence message


def cmf_aoadmm(
    matrices,
    rank,
    init="random",
    n_iter_max=1000,
    l2_penalty=None,
    tv_penalty=None,
    l1_penalty=None,
    non_negative=None,
    unimodal=None,
    generalized_l2_penalty=None,
    l2_norm_bound=None,
    lower_bound=None,
    upper_bound=None,
    parafac2=None,
    regs=None,
    feasibility_penalty_scale=1,
    constant_feasibility_penalty=False,
    aux_init="random_uniform",
    dual_init="random_uniform",
    svd="truncated_svd",
    init_params=None,
    random_state=None,
    tol=1e-8,
    absolute_tol=1e-10,
    feasibility_tol=1e-4,
    inner_tol=None,
    inner_n_iter_max=5,
    update_A=True,
    update_B_is=True,
    update_C=True,
    return_admm_vars=False,
    return_errors=False,
    verbose=False,
    mask=None,
):
    r"""Fit a regularized coupled matrix factorization model with AO-ADMM

    Regularization parameters can be:

     * ``None`` (no regularization)
     * A single regularization parameter (same regularization in all modes)
     * A list with specific regularization parameters (one for each mode)
     * A dictionary with mode-index (0, 1 or 2) as keys and regularization parameter as values
       (regularization only in specific modes)

    :math:`\mathbf{A}` is represented by mode-index 0, :math:`\{\mathbf{B}^{(i)}\}_{i=1}^I` is
    represented by mode-index 1 and :math:`\mathbf{C}` is represented by mode-index 2.

    Parameters
    ----------
    matrices : list of ndarray
        Matrices that the coupled matrix factorization should model
    rank : int
        The rank of the model
    init : {"random", "svd", "threshold_svd", "parafac_als", "parafac_hals", "parafac2_als"} (default="random")
        Initialization method. If equal to ``"parafac_als"``, ``"parafac_hals"`` or ``"parafac2_als"``,
        then the corresponding methods in TensorLy are used to initialize the model. In that case,
        you can pass additional keyword-arguments to the decomposition function with the ``init_params``
        parameter.
    n_iter_max : int (default=1000)
        Maximum number of iterations.
    l2_penalty : Regularization parameter (default=None)
        Strength of the L2 penalty, imposed as ``0.5 * l2_penalty * tl.sum(M**2)``, where ``M``
        represents a single factor matrix (note that this differs by a factor :math:`0.5` compared
        to the expression in :cite:p:`roald2021parafac2,roald2021admm`).
    tv_penalty : Regularization parameter (default=None)
        Strength of the TV penalty. To use this regularizer, you must have the GPL-lisenced library:
        ``condat_tv`` installed.
    l1_penalty : Regularization parameter (default=None)
        Strength of the sparsity inducing regularization
    non_negative : Regularization parameter (default=None)
        If True, then the corresponding factor matrices are non-negative
    unimodal : Regularization parameter (default=None)
        If True, then the corresponding factor matrices have unimodal columns
    generalized_l2_penalty : Regularization parameter (must be ``None``, list or dict) (default=None)
        List or dict containing square matrices that parametrize a generalized L2 norm.
    l2_norm_bound : Regularization parameter (default=None)
        Maximum L2 norm of the columns
    lower_bound : Regularization parameter (default=None)
        Lower value of the columns
    upper_bound : Regularization parameter (default=None)
        Upper value of the columns
    parafac2 : bool (default=False)
        If ``True``, then the :math:`\mathbf{B}^{(i)}`-matrices follow the PARAFAC2 constraint
    regs : List of lists of matcouply.penalties.ADMMPenalty (optional, default=None)
        The first element of this list contains a list of ``matcouply.penalties.ADMMPenalty``-instances
        for :math:`\mathbf{A}`, the second for :math:`\{\mathbf{B}^{(i)}\}_{i=1}^I` and the third for
        :math:`\mathbf{C}`.
    feasibility_penalty_scale : float (default=1)
        What function to multiply the automatically computed feasibility penalty parameter by
        (see :cite:p:`roald2021admm` for details)
    constant_feasibility_penalty : bool, "A" or "B" (default=False)
        If True, then all rows of :math:`\mathbf{A}` will have the same feasibility penalty parameter
        and all :math:`\mathbf{B}^{(i)}`-matrices will have the same feasibility penalty parameter.
        This makes it possible to use matrix-penalties for :math:`\mathbf{A}` and multi-matrix penalties
        that require the same penalty parameter for all :math:`\mathbf{B}^{(i)}`-matrices.

        The maximum penalty parameter for all rows of :math:`\mathbf{A}` are used as the penalty
        parameter for :math:`\mathbf{A}` and the maximum penalty parameter among all :math:`\mathbf{B}^{(i)}`-matrices
        are used as the penalty parameter for :math:`\{\mathbf{B}^{(i)}\}_{i=1}^I`.

        If ``"A"`` or ``"B"``, then this option is only enabled for the specified parameters.
    aux_init : str (default="random_uniform")
        Initialization scheme for the auxiliary variables. See :meth:`matcouply.penalties.ADMMPenalty.init_aux`
        for possible options.
    dual_init : str (default="random_uniform")
        Initialization scheme for the dual variables. See :meth:`matcouply.penalties.ADMMPenalty.init_dual`
        for possible options.
    svd : str (default="truncated_svd")
        String that specifies which SVD algorithm to use. Valid strings are the keys of ``tensorly.SVD_FUNS``.
    init_params : dict
        Keyword arguments to use for the initialization.
    random_state : None, int or valid tensorly random state
    tol : float (default=1e-8)
        Relative loss decrease condition. For stopping, the optimization requires
        that ``abs(losses[-2] - losses[-1]) / losses[-2] < tol`` or that
        ``losses[-2] < absolute_tol``.
    absolute_tol : float (default=1e-10)
        Absolute loss value condition. For stopping, the optimization requires
        that ``abs(losses[-2] - losses[-1]) / losses[-2] < tol`` or that
        ``losses[-2] < absolute_tol``.
    feasibility_tol : float (default=1e-4)
        If any feasibility gap is greater than feasibility_tol, then the optimization
        will not stop (unless the maximum number of iterations are reached).
    inner_tol : float (optional, default=None)
        If set, then this specifies the stopping condition for the inner subproblems,
        solved with ADMM. If it is None, then the ADMM algorithm will always run for
        ``inner_n_iter_max``-iterations.
    inner_n_iter_max : int (default=5)
        Maximum number of iterations for the ADMM subproblems
    update_A : bool (default=True)
        If ``False``, then :math:`\mathbf{A}` will not be updated.
    update_B_is : bool (default=True)
        If ``False``, then the :math:`\mathbf{B}^{(i)}`-matrices will not be updated.
    update_C : bool (default=True)
        If ``False``, then :math:`\mathbf{C}` will not be updated.
    return_admm_vars : bool (default=False)
        If ``True``, then the auxiliary and dual variables will also be returned.
    return_errors : bool (default=False)
        If ``True``, then the reconstruction error, feasibility gaps and loss per iteration
        will be returned.
    verbose : bool or int (default=False)
        If ``True``, then a message with convergence info will be printed out every iteration.
        If ``int > 1``, then a message with convergence info will be printed out ever ``verbose`` iteration.
    mask : ndarray, optional
        An array with the same shape as the tensor. It should be 0 where there are missing values and 1 everywhere else.

    Returns
    -------
    CoupledMatrixFactorization
        The fitted coupled matrix factorization
    ADMMVars
        (only returned if ``return_admm_vars=True``) NamedTuple containing the auxuliary
        and dual variables. The feasibility gaps are computed by taking the relative normed
        differences between the auxiliary variables and the factor matrices of the model.
    DiagnosticMetrics
        (only returned if ``return_errors=True``) NamedTuple containing lists of the relative
        norm-error, the feasibility gaps and the regularized loss for each iteration, each with
        length equal to the number of iterations plus one (the initial values), a stopping
        message and two boolean values, one signifying whether the stopping conditions (including
        feasibility gap) were satisfied and one signifying whether the feasibility gaps were
        sufficiently low (according to ``feasibility_tol``).

    Note
    ----
    If the maximum number of iterations is reached, then you should check that the feasibility
    gaps are sufficiently small (see :doc:`../optimization` for more details)

    Note
    ----
    If you use norm-dependent regularization (e.g. ``generalized_l2_penalty``) in one mode,
    then you must use norm-dependent regularization in all modes. You may, for example, use
    ``l2_penalty``, ``norm_bound``, ``l1_penalty`` or ``lower_bound`` and ``upper_bound``.
    See :cite:p:`roald2021admm` for more details.

    Note
    ----
    For simplicity, the model used here is a permuted version of that in
    :cite:p:`roald2021parafac2,roald2021admm` (where :math:`\mathbf{B_k}`-matrices
    vary over the rows in :math:`\mathbf{C}` instead of the rows in :math:`\mathbf{A}`).

    Examples
    --------
    Here is a small example that shows what the diagnostic metrics contain. For more detailed examples, see :ref:`examples`

    >>> import matcouply
    >>> data = matcouply.random.random_coupled_matrices([(10, 5), (12, 5), (11, 5)], rank=2, full=True, random_state=1)
    >>> cmf, diagnostics = matcouply.decomposition.cmf_aoadmm(data, 2, n_iter_max=4, non_negative=True, return_errors=True, random_state=10)
    >>> diagnostics.message
    'MAXIMUM NUMBER OF ITERATIONS REACHED'

    >>> diagnostics.satisfied_stopping_condition
    False

    >>> diagnostics.satisfied_feasibility_condition
    np.True_

    >>> len(diagnostics.regularized_loss)
    5

    We see that the length of the regularized loss list is the number of iterations plus one. This is because the initial
    decomposition is included in this.

    >>> len(diagnostics.feasibility_gaps[0])
    3

    The feasibility gaps list contain tuples (one for each mode) of tuples (one for each penalty for the given mode).

    >>> len(diagnostics.feasibility_gaps[0][0])
    1
    """
    random_state = tl.check_random_state(random_state)
    svd_fun = get_svd(svd)
    cmf = initialize_cmf(matrices, rank, init, svd_fun=svd_fun, random_state=random_state, init_params=init_params)

    # Parse constraints
    l2_penalty = _listify(l2_penalty, "l2_penalty")
    l2_penalty = [l2 if l2 is not None else 0 for l2 in l2_penalty]

    # Initial missing imputation
    if mask is not None:
        if init == "random" or init == "svd" or init == "threshold_svd":
            matrices = _update_imputed(tensor_slices=list(matrices), mask=mask, decomposition=None, method="mode-3")
        else:  # If factors are provided from a "warmer" start (e.g. parafac2_als) use the factor estimates as initial guesses
            matrices = _update_imputed(tensor_slices=list(matrices), mask=mask, decomposition=cmf, method="factors")

    regs = _parse_all_penalties(
        non_negative=non_negative,
        lower_bound=lower_bound,
        upper_bound=upper_bound,
        l2_norm_bound=l2_norm_bound,
        unimodal=unimodal,
        parafac2=parafac2,
        l1_penalty=l1_penalty,
        tv_penalty=tv_penalty,
        generalized_l2_penalty=generalized_l2_penalty,
        svd=svd,
        regs=regs,
        dual_init=dual_init,
        aux_init=aux_init,
        verbose=verbose,
    )

    if not update_A:
        regs[0] = []
    if not update_B_is:
        regs[1] = []
    if not update_C:
        regs[2] = []

    # TODO  (Improvement): Include cmf to initialize functions in case other init schemes require that?
    (
        A_aux_list,
        B_is_aux_list,
        C_aux_list,
    ) = initialize_aux(matrices, rank, regs, random_state=random_state)
    (
        A_dual_list,
        B_is_dual_list,
        C_dual_list,
    ) = initialize_dual(matrices, rank, regs, random_state=random_state)
    norm_matrices = _root_sum_squared_list(matrices, mask=mask)
    rec_errors = []
    feasibility_gaps = []

    rec_error = _cmf_reconstruction_error(matrices=matrices, cmf=cmf, norm_matrices=norm_matrices,intermediate_A_calculations=None,mask=mask)
    rec_error /= norm_matrices
    rec_errors.append(rec_error)
    losses = []
    reg_penalty = (
        +sum(A_reg.penalty(cmf[1][0]) for A_reg in regs[0])
        + sum(B_reg.penalty(cmf[1][1]) for B_reg in regs[1])
        + sum(C_reg.penalty(cmf[1][2]) for C_reg in regs[2])
    )
    l2_reg = _compute_l2_penalty(cmf, l2_penalty)
    losses.append(0.5 * rec_error**2 + l2_reg + reg_penalty)

    A_gaps, B_gaps, C_gaps = compute_feasibility_gaps(cmf, regs, A_aux_list, B_is_aux_list, C_aux_list)
    feasibility_gaps.append((A_gaps, B_gaps, C_gaps))
    if verbose and verbose > 0:
        print("Feasibility gaps for A: {}".format(A_gaps))
        print("Feasibility gaps for the Bi-matrices: {}".format(B_gaps))
        print("Feasibility gaps for C: {}".format(C_gaps))

    # Default values for diagnostics
    satisfied_stopping_condition = False
    message = "MAXIMUM NUMBER OF ITERATIONS REACHED"
    if isinstance(constant_feasibility_penalty, str) and constant_feasibility_penalty not in {"A", "B"}:
        raise ValueError(
            f"If `constant_feasibility_penalty` is a string, it must be 'A' or 'B', not {constant_feasibility_penalty}"
        )
    constant_A_feasibility = (
        constant_feasibility_penalty and not isinstance(constant_feasibility_penalty, str)
    ) or constant_feasibility_penalty == "A"
    constant_B_feasibility = (
        constant_feasibility_penalty and not isinstance(constant_feasibility_penalty, str)
    ) or constant_feasibility_penalty == "B"

    it = -1  # Needed if n_iter_max <= 0
    for it in range(n_iter_max):
        intermediate_A_calculations = None

        if update_B_is:
            cmf, B_is_aux_list, B_is_dual_list = admm_update_B(
                matrices=matrices,
                reg=regs[1],
                cmf=cmf,
                B_is_aux_list=B_is_aux_list,
                B_is_dual_list=B_is_dual_list,
                l2_penalty=l2_penalty[1],
                inner_n_iter_max=inner_n_iter_max,
                inner_tol=inner_tol,
                feasibility_penalty_scale=feasibility_penalty_scale,
                constant_feasibility_penalty=constant_B_feasibility,
                svd_fun=svd_fun,
            )
        if update_C:
            cmf, C_aux_list, C_dual_list = admm_update_C(
                matrices=matrices,
                reg=regs[2],
                cmf=cmf,
                C_aux_list=C_aux_list,
                C_dual_list=C_dual_list,
                l2_penalty=l2_penalty[2],
                inner_n_iter_max=inner_n_iter_max,
                inner_tol=inner_tol,
                feasibility_penalty_scale=feasibility_penalty_scale,
                svd_fun=svd_fun,
            )
        if update_A:
            cmf, A_aux_list, A_dual_list, intermediate_A_calculations = admm_update_A(
                matrices=matrices,
                reg=regs[0],
                cmf=cmf,
                A_aux_list=A_aux_list,
                A_dual_list=A_dual_list,
                l2_penalty=l2_penalty[0],
                inner_n_iter_max=inner_n_iter_max,
                inner_tol=inner_tol,
                feasibility_penalty_scale=feasibility_penalty_scale,
                constant_feasibility_penalty=constant_A_feasibility,
                svd_fun=svd_fun,
            )

        if mask is not None:

            matrices = _update_imputed(tensor_slices=matrices, decomposition=cmf, mask=mask, method="factors")

        if tol or absolute_tol or return_errors:
            curr_feasibility_gaps = compute_feasibility_gaps(cmf, regs, A_aux_list, B_is_aux_list, C_aux_list)
            feasibility_gaps.append(curr_feasibility_gaps)

            if tol or absolute_tol:
                # Compute stopping criterions
                feasibility_criterion = feasibility_tol and _check_feasibility(curr_feasibility_gaps, feasibility_tol)

                if not feasibility_criterion and not return_errors:
                    A_gaps, B_gaps, C_gaps = curr_feasibility_gaps
                    if verbose and it % verbose == 0 and verbose > 0:
                        print(
                            "Coupled matrix factorization iteration={}, ".format(it)
                            + "reconstruction error=NOT COMPUTED, "
                            + "regularized loss=NOT COMPUTED, "
                            + "regularized loss variation=NOT COMPUTED."
                        )
                        print("Feasibility gaps for A: {}".format(A_gaps))
                        print("Feasibility gaps for the Bi-matrices: {}".format(B_gaps))
                        print("Feasibility gaps for C: {}".format(C_gaps))

                    continue

            rec_error = _cmf_reconstruction_error(matrices=matrices,cmf=cmf,norm_matrices=norm_matrices,intermediate_A_calculations=intermediate_A_calculations,mask=mask)
            rec_error /= norm_matrices
            rec_errors.append(rec_error)
            reg_penalty = (
                +sum(A_reg.penalty(cmf[1][0]) for A_reg in regs[0])
                + sum(B_reg.penalty(cmf[1][1]) for B_reg in regs[1])
                + sum(C_reg.penalty(cmf[1][2]) for C_reg in regs[2])
            )

            l2_reg = _compute_l2_penalty(cmf, l2_penalty)
            losses.append(0.5 * rec_error**2 + l2_reg + reg_penalty)

            if verbose and it % verbose == 0 and verbose > 0:
                A_gaps, B_gaps, C_gaps = curr_feasibility_gaps
                print(
                    "Coupled matrix factorization iteration={}, ".format(it)
                    + "reconstruction error={}, ".format(rec_errors[-1])
                    + "regularized loss={} ".format(losses[-1])
                    + "regularized loss variation={}.".format(abs(losses[-2] - losses[-1]) / losses[-2])
                )
                print("Feasibility gaps for A: {}".format(A_gaps))
                print("Feasibility gaps for the Bi-matrices: {}".format(B_gaps))
                print("Feasibility gaps for C: {}".format(C_gaps))

            if tol:
                # Compute rest of stopping criterions
                rel_loss_criterion = abs(losses[-2] - losses[-1]) < (tol * losses[-2])
                abs_loss_criterion = losses[-1] < absolute_tol

                if feasibility_criterion and rel_loss_criterion:
                    satisfied_stopping_condition = True
                    message = "FEASIBILITY GAP CRITERION AND RELATIVE LOSS CRITERION SATISFIED"
                    if verbose:
                        print("converged in {} iterations: {}".format(it, message))
                    break
                elif feasibility_criterion and abs_loss_criterion:
                    satisfied_stopping_condition = True
                    message = "FEASIBILITY GAP CRITERION AND ABSOLUTE LOSS CRITERION SATISFIED"
                    if verbose:
                        print("converged in {} iterations: {}".format(it, message))
                    break

        elif verbose and it % verbose == 0 and verbose > 0:
            print("Coupled matrix factorization iteration={}".format(it))
    else:
        if verbose:
            print("REACHED MAXIMUM NUMBER OF ITERATIONS")

    # Compute feasibility gaps to return with diagnostics
    # If the feasibility tolerance is set, but no loss tolerance then the feasibility criterion is not
    # computed in the AO-ADMM loop. Likewise, if n_iter_max <= 0, the feasibility tolerance
    # is not computed. This is only relevant if we should return errors, but it is a fast operation
    # so we do it anyways.
    if feasibility_tol and return_errors:
        curr_feasibility_gaps = compute_feasibility_gaps(cmf, regs, A_aux_list, B_is_aux_list, C_aux_list)

        feasibility_criterion = _check_feasibility(curr_feasibility_gaps, feasibility_tol)
    elif not feasibility_tol:
        feasibility_criterion = None

    # Save as validated factorization instead of tuple
    cmf = CoupledMatrixFactorization(cmf)

    out = [cmf]
    if return_admm_vars:
        admm_vars = ADMMVars(
            auxes=(A_aux_list, B_is_aux_list, C_aux_list), duals=(A_dual_list, B_is_dual_list, C_dual_list)
        )
        out.append(admm_vars)
    if return_errors:
        if not satisfied_stopping_condition and not (tol or absolute_tol):
            satisfied_stopping_condition = None

        diagnostic_metrics = DiagnosticMetrics(
            rec_errors=rec_errors,
            feasibility_gaps=feasibility_gaps,
            regularized_loss=losses,
            satisfied_stopping_condition=satisfied_stopping_condition,
            satisfied_feasibility_condition=feasibility_criterion,
            message=message,
            n_iter=it + 1,  # Plus one since this is the number of iterations, not the iteration number
        )
        out.append(diagnostic_metrics)

    if len(out) == 1:
        return out[0]
    else:
        return tuple(out)


def parafac2_aoadmm(
    matrices,
    rank,
    init="random",
    n_iter_max=1000,
    l2_penalty=0,
    tv_penalty=None,
    l1_penalty=None,
    non_negative=None,
    unimodal=None,
    generalized_l2_penalty=None,
    l2_norm_bound=None,
    lower_bound=None,
    upper_bound=None,
    regs=None,
    feasibility_penalty_scale=1,
    constant_feasibility_penalty=False,
    aux_init="random_uniform",
    dual_init="random_uniform",
    svd="truncated_svd",
    init_params=None,
    random_state=None,
    tol=1e-8,
    absolute_tol=1e-10,
    feasibility_tol=1e-4,
    inner_tol=None,
    inner_n_iter_max=5,
    update_A=True,
    update_B_is=True,
    update_C=True,
    return_errors=False,
    return_admm_vars=False,
    verbose=False,
    mask=None,
):
    """Alias for cmf_aoadmm with PARAFAC2 constraint (constant cross-product) on mode 1 (B mode)

    See also
    --------
    matcouply.decomposition.cmf_aoadmm : General coupled matrix factorization with AO-ADMM
    matcouply.penalties.Parafac2 : Class for PARAFAC2 constraint with more information about its properties
    """

    return cmf_aoadmm(
        matrices=matrices,
        rank=rank,
        init=init,
        n_iter_max=n_iter_max,
        l2_penalty=l2_penalty,
        tv_penalty=tv_penalty,
        l1_penalty=l1_penalty,
        non_negative=non_negative,
        unimodal=unimodal,
        generalized_l2_penalty=generalized_l2_penalty,
        l2_norm_bound=l2_norm_bound,
        lower_bound=lower_bound,
        upper_bound=upper_bound,
        parafac2=True,
        regs=regs,
        feasibility_penalty_scale=feasibility_penalty_scale,
        constant_feasibility_penalty=constant_feasibility_penalty,
        aux_init=aux_init,
        dual_init=dual_init,
        svd=svd,
        init_params=init_params,
        random_state=random_state,
        tol=tol,
        absolute_tol=absolute_tol,
        feasibility_tol=feasibility_tol,
        inner_tol=inner_tol,
        inner_n_iter_max=inner_n_iter_max,
        update_A=update_A,
        update_B_is=update_B_is,
        update_C=update_C,
        return_errors=return_errors,
        return_admm_vars=return_admm_vars,
        verbose=verbose,
        mask=mask,
    )



# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

from abc import ABCMeta
from functools import partial


# mro function by Steven D'Aprano (2011). Released under an MIT lisence. https://code.activestate.com/recipes/577748-calculate-the-mro-of-a-class/
def mro(*bases):
    """Calculate the Method Resolution Order of bases using the C3 algorithm.

    Suppose you intended creating a class K with the given base classes. This
    function returns the MRO which K would have, *excluding* K itself (since
    it doesn't yet exist), as if you had actually created the class.

    Another way of looking at this, if you pass a single class K, this will
    return the linearization of K (the MRO of K, *including* itself).
    """
    seqs = [list(C.__mro__) for C in bases] + [list(bases)]
    res = []
    while True:
        non_empty = list(filter(None, seqs))
        if not non_empty:
            # Nothing left to process, we're done.
            return tuple(res)
        for seq in non_empty:  # Find merge candidates among seq heads.
            candidate = seq[0]
            not_head = [s for s in non_empty if candidate in s[1:]]
            if not_head:
                # Reject the candidate.
                candidate = None
            else:
                break
        if not candidate:
            raise TypeError("inconsistent hierarchy, no C3 MRO is possible")
        res.append(candidate)
        for seq in non_empty:
            # Remove candidate.
            if seq[0] == candidate:
                del seq[0]


# Code below by nikratio (2013). Released under an MIT lisence. https://code.activestate.com/recipes/578587-inherit-method-docstrings-without-breaking-decorat/
# The code is modified slightly to allow for metaclasses
# This definition is only used to assist static code analyzers
def copy_ancestor_docstring(fn):
    """Copy docstring for method from superclass

    For this decorator to work, the class has to use the `InheritableDocstrings`
    metaclass.
    """
    raise RuntimeError("Decorator can only be used in classes " "using the `InheritableDocstrings` metaclass")


def _copy_ancestor_docstring(mro, fn):
    """Decorator to set docstring for *fn* from *mro*"""

    if fn.__doc__ is not None:
        raise RuntimeError("Function already has docstring")

    # Search for docstring in superclass
    for cls in mro:
        super_fn = getattr(cls, fn.__name__, None)
        if super_fn is None:
            continue
        fn.__doc__ = super_fn.__doc__
        break
    else:
        raise RuntimeError("Can't inherit docstring for %s: method does not " "exist in superclass" % fn.__name__)

    return fn


class InheritableDocstrings(ABCMeta):
    @classmethod
    def __prepare__(cls, name, bases, **kwds):
        classdict = super().__prepare__(name, bases, *kwds)

        # Inject decorators into class namespace
        classdict["copy_ancestor_docstring"] = partial(_copy_ancestor_docstring, mro(*bases))

        return classdict

    def __new__(cls, name, bases, classdict):

        # Decorator may not exist in class dict if the class (metaclass
        # instance) was constructed with an explicit call to `type`.
        # (cf http://bugs.python.org/issue18334)
        if "copy_ancestor_docstring" in classdict:

            # Make sure that class definition hasn't messed with decorators
            copy_impl = getattr(classdict["copy_ancestor_docstring"], "func", None)
            if copy_impl is not _copy_ancestor_docstring:
                raise RuntimeError(
                    "No copy_ancestor_docstring attribute may be created "
                    "in classes using the InheritableDocstrings metaclass"
                )

            # Delete decorators from class namespace
            del classdict["copy_ancestor_docstring"]

        return super().__new__(cls, name, bases, classdict)



# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

import pytest
import tensorly as tl
from numpy.linalg import matrix_rank

from matcouply.random import random_coupled_matrices
from matcouply.testing import assert_allclose


def test_random_coupled_matrices():
    """test for random.random_coupled_matrices"""
    shapes = [(10, 11), (12, 11), (9, 11), (10, 11), (15, 11)]
    rank = 4

    # Check that assembled matrices have correct shapes and ranks
    coupled_matrices = random_coupled_matrices(shapes, rank, full=True)

    assert len(coupled_matrices) == len(shapes)
    for matrix, shape in zip(coupled_matrices, shapes):
        assert tl.shape(matrix) == shape
        array = tl.to_numpy(matrix)
        assert matrix_rank(array) == rank

    # Check that factor matrices have correct shape
    weights, (A, B_is, C) = random_coupled_matrices(shapes, rank, full=False)
    assert tl.shape(A) == (len(shapes), rank)
    assert all(tl.shape(B_i) == (J_i, rank) for B_i, (J_i, K) in zip(B_is, shapes))
    assert tl.shape(C) == (shapes[0][1], rank)

    # Check that normalising B_is gives B_is with unit normed columns
    weights, (A, B_is, C) = random_coupled_matrices(shapes, rank, full=False, normalise_B=True, normalise_factors=False)
    for B_i in B_is:
        assert_allclose(tl.norm(B_i, axis=0), tl.ones(rank), rtol=1e-6)  # rtol=1e-6 due to PyTorch single precision
    assert_allclose(weights, tl.ones(rank), rtol=1e-6)  # rtol=1e-6 due to PyTorch single precision

    # Check that normalising all factors gives factor matrices with unit normed columns
    weights, (A, B_is, C) = random_coupled_matrices(shapes, rank, full=False, normalise_factors=True)
    assert_allclose(tl.norm(A, axis=0), tl.ones(rank), rtol=1e-6)  # rtol=1e-6 due to PyTorch single precision
    for B_i in B_is:
        assert_allclose(tl.norm(B_i, axis=0), tl.ones(rank), rtol=1e-6)  # rtol=1e-6 due to PyTorch single precision
    assert_allclose(tl.norm(C, axis=0), tl.ones(rank), rtol=1e-6)  # rtol=1e-6 due to PyTorch single precision

    # Should fail when shapes have different value for the number of columns
    shapes = [(10, 10), (12, 11), (9, 11), (10, 11), (15, 11)]
    with pytest.raises(ValueError):
        random_coupled_matrices(shapes, rank)



# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

import math
from copy import copy
from unittest.mock import patch

import numpy as np
import pytest
import scipy.stats as stats
import tensorly as tl
from tensorly.testing import assert_array_equal

from matcouply import penalties
from matcouply._utils import get_svd
from matcouply.testing import (
    BaseTestFactorMatricesPenalty,
    BaseTestFactorMatrixPenalty,
    BaseTestRowVectorPenalty,
    MixinTestHardConstraint,
    assert_allclose,
)

from .utils import RTOL_SCALE  # TODO: Consider if this should be changed


def test_row_vector_penalty_forwards_updates_correctly(rng, random_matrix, random_matrices):
    class BetweenZeroAndOneConstraint(penalties.RowVectorPenalty):
        def factor_matrix_row_update(self, factor_matrix_row, feasibility_penalty, aux_row):
            return tl.clip(factor_matrix_row, 0, 1)

        def penalty(self, x):
            return 0

    penalty = BetweenZeroAndOneConstraint()
    updated_factor_matrix = penalty.factor_matrix_update(random_matrix, 1, random_matrix)  # last input is ignored

    assert tl.all(updated_factor_matrix >= 0)
    assert tl.all(updated_factor_matrix <= 1)

    updated_factor_matrices = penalty.factor_matrices_update(
        random_matrices, np.ones(len(random_matrices)), random_matrices
    )  # last input is ignored
    for factor_matrix in updated_factor_matrices:
        assert tl.all(factor_matrix >= 0)
        assert tl.all(factor_matrix <= 1)


def test_matrix_penalty_forwards_updates_correctly(rng, random_matrices):
    class BetweenZeroAndOneConstraint(penalties.MatrixPenalty):
        def factor_matrix_update(self, factor_matrix, feasibility_penalty, aux_row):
            return tl.clip(factor_matrix, 0, 1)

        def penalty(self, x):
            return 0

    penalty = BetweenZeroAndOneConstraint()
    updated_factor_matrices = penalty.factor_matrices_update(
        random_matrices, np.ones(len(random_matrices)), random_matrices
    )  # last input is ignored
    for factor_matrix in updated_factor_matrices:
        assert tl.all(factor_matrix >= 0)
        assert tl.all(factor_matrix <= 1)


class TestL1Penalty(BaseTestRowVectorPenalty):
    PenaltyType = penalties.L1Penalty
    penalty_default_kwargs = {"reg_strength": 1}

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_row_update_invariant_point(self, non_negativity):
        invariant_matrix_row = tl.zeros((1, 4))
        l1_penalty = penalties.L1Penalty(0.1, non_negativity=non_negativity)

        out = l1_penalty.factor_matrix_row_update(invariant_matrix_row, 10, None)
        assert_allclose(invariant_matrix_row, out)

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_factor_matrix_update_invariant_point(self, non_negativity):
        invariant_matrix = tl.zeros((10, 3))
        l1_penalty = penalties.L1Penalty(0.1, non_negativity=non_negativity)

        out = l1_penalty.factor_matrix_update(invariant_matrix, 10, None)
        assert_allclose(invariant_matrix, out)

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_factor_matrices_update_invariant_point(self, non_negativity):
        invariant_matrices = [tl.zeros((10, 3)) for i in range(5)]
        feasibility_penalties = [10] * len(invariant_matrices)
        auxes = [None] * len(invariant_matrices)
        l1_penalty = penalties.L1Penalty(0.1, non_negativity=non_negativity)

        out = l1_penalty.factor_matrices_update(invariant_matrices, feasibility_penalties, auxes)
        for invariant_matrix, out_matrix in zip(invariant_matrices, out):
            assert_allclose(invariant_matrix, out_matrix)

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_row_update_reduces_penalty(self, random_row, non_negativity):
        l1_penalty = penalties.L1Penalty(0.1, non_negativity=non_negativity)

        initial_penalty = l1_penalty.penalty(random_row)
        out = l1_penalty.factor_matrix_row_update(random_row, 10, None)
        assert l1_penalty.penalty(out) <= initial_penalty

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_factor_matrix_update_reduces_penalty(self, random_matrix, non_negativity):
        l1_penalty = penalties.L1Penalty(0.1, non_negativity=non_negativity)

        initial_penalty = l1_penalty.penalty(random_matrix)
        out = l1_penalty.factor_matrix_update(random_matrix, 10, None)
        assert l1_penalty.penalty(out) <= initial_penalty

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_factor_matrices_update_reduces_penalty(self, random_matrices, non_negativity):
        l1_penalty = penalties.L1Penalty(0.1, non_negativity=non_negativity)
        feasibility_penalties = [10] * len(random_matrices)
        auxes = [None] * len(random_matrices)
        initial_penalty = l1_penalty.penalty(random_matrices)
        out = l1_penalty.factor_matrices_update(random_matrices, feasibility_penalties, auxes)
        assert l1_penalty.penalty(out) <= initial_penalty

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_row_update_changes_input(self, random_row, non_negativity):
        l1_penalty = penalties.L1Penalty(0.1, non_negativity=non_negativity)

        out = l1_penalty.factor_matrix_row_update(random_row, 10, None)
        assert not np.allclose(out, random_row)

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_factor_matrix_update_changes_input(self, random_matrix, non_negativity):
        l1_penalty = penalties.L1Penalty(0.1, non_negativity=non_negativity)

        out = l1_penalty.factor_matrix_update(random_matrix, 10, None)
        assert not np.allclose(out, random_matrix)

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_factor_matrices_update_changes_input(self, random_matrices, non_negativity):
        feasibility_penalties = [10] * len(random_matrices)
        auxes = [None] * len(random_matrices)
        l1_penalty = penalties.L1Penalty(0.1, non_negativity=non_negativity)

        out = l1_penalty.factor_matrices_update(random_matrices, feasibility_penalties, auxes)
        for random_matrix, out_matrix in zip(random_matrices, out):
            assert not np.allclose(random_matrix, out_matrix)

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_factor_matrix_update_sets_small_weights_to_zero(self, random_matrix, non_negativity):
        random_matrix /= tl.abs(random_matrix).max()
        feasibility_penalty = 1
        aux = None
        l1_penalty = penalties.L1Penalty(1, non_negativity=non_negativity)

        out = l1_penalty.factor_matrix_update(random_matrix, feasibility_penalty, aux)
        assert_allclose(out, 0)

    def test_non_negativity_sets_negative_values_to_zero(self):
        negative_matrix = tl.ones((30, 5)) * (-100)
        feasibility_penalty = 1
        aux = None
        l1_penalty = penalties.L1Penalty(1, non_negativity=True)

        out = l1_penalty.factor_matrix_update(negative_matrix, feasibility_penalty, aux)
        assert_array_equal(out, 0)

    def test_penalty(self, random_ragged_cmf):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        l1_penalty = self.PenaltyType(reg_strength=1)
        assert l1_penalty.penalty(A) == pytest.approx(tl.sum(tl.abs(A)))
        assert l1_penalty.penalty(B_is) == pytest.approx(sum(tl.sum(tl.abs(B_i)) for B_i in B_is))
        assert l1_penalty.penalty(C) == pytest.approx(tl.sum(tl.abs(C)))

        l1_penalty = self.PenaltyType(reg_strength=2)
        assert l1_penalty.penalty(A) == pytest.approx(2 * tl.sum(tl.abs(A)))
        assert l1_penalty.penalty(B_is) == pytest.approx(2 * sum(tl.sum(tl.abs(B_i)) for B_i in B_is))
        assert l1_penalty.penalty(C) == pytest.approx(2 * tl.sum(tl.abs(C)))

        l1_penalty = self.PenaltyType(reg_strength=0)
        assert l1_penalty.penalty(A) == 0
        assert l1_penalty.penalty(B_is) == 0
        assert l1_penalty.penalty(C) == 0

        with pytest.raises(ValueError):
            l1_penalty = self.PenaltyType(reg_strength=-1)


class TestBox(MixinTestHardConstraint, BaseTestRowVectorPenalty):
    PenaltyType = penalties.Box
    penalty_default_kwargs = {"min_val": 0, "max_val": 1}

    def get_invariant_row(self, rng, n_columns):
        invariant_row = tl.tensor(rng.uniform(size=(1, n_columns), low=0, high=1))
        return invariant_row

    def get_non_invariant_row(self, rng, n_columns):
        random_row = tl.tensor(rng.uniform(size=(1, n_columns)))
        random_row[0] += 100
        return random_row


class TestL2BallConstraint(MixinTestHardConstraint, BaseTestFactorMatrixPenalty):
    PenaltyType = penalties.L2Ball
    penalty_default_kwargs = {"norm_bound": 1}

    def get_invariant_matrix(self, rng, shape):
        random_matrix = tl.tensor(rng.random_sample(shape))
        return random_matrix / math.sqrt(shape[0])

    def get_non_invariant_matrix(self, rng, shape):
        random_matrix = tl.tensor(rng.random_sample(shape) + 10)
        return 10 + random_matrix / math.sqrt(shape[0])

    def test_input_is_checked(self):
        with pytest.raises(ValueError):
            self.PenaltyType(norm_bound=0)
        with pytest.raises(ValueError):
            self.PenaltyType(norm_bound=-1)

        self.PenaltyType(norm_bound=0.1)

    def test_non_negativity_sets_negative_values_to_zero(self):
        negative_matrix = tl.ones((30, 5)) * (-100)
        feasibility_penalty = 1
        aux = None
        penalty = self.PenaltyType(1, non_negativity=True)

        out = penalty.factor_matrix_update(negative_matrix, feasibility_penalty, aux)
        assert_array_equal(out, 0)


class TestUnitSimplex(MixinTestHardConstraint, BaseTestFactorMatrixPenalty):
    PenaltyType = penalties.UnitSimplex
    rtol = 1e-5 * RTOL_SCALE
    atol = 1e-8

    def get_invariant_matrix(self, rng, shape):
        random_matrix = tl.tensor(rng.uniform(size=shape))
        col_sum = tl.sum(random_matrix, axis=0)
        return random_matrix / col_sum

    def get_non_invariant_matrix(self, rng, shape):
        random_matrix = tl.tensor(rng.uniform(size=shape))
        col_sum = tl.sum(random_matrix, axis=0)
        return 10 + random_matrix / col_sum


class TestGeneralizedL2Penalty(BaseTestFactorMatrixPenalty):
    PenaltyType = penalties.GeneralizedL2Penalty
    n_rows = 50
    min_rows = n_rows
    max_rows = n_rows

    norm_matrix1 = 2 * np.eye(n_rows // 2) - np.eye(n_rows // 2, k=-1) - np.eye(n_rows // 2, k=1)
    norm_matrix1[0, 0] = 1
    norm_matrix1[-1, -1] = 1

    norm_matrix2 = 2 * np.eye(n_rows // 2) - np.eye(n_rows // 2, k=-1) - np.eye(n_rows // 2, k=1)
    norm_matrix2[0, 0] = 1
    norm_matrix2[-1, -1] = 1

    zeros_matrix = np.zeros((n_rows // 2, n_rows // 2))

    # fmt: off
    norm_matrix = tl.tensor(np.block([
        [norm_matrix1, zeros_matrix],
        [zeros_matrix, norm_matrix2]
    ]))
    # fmt: on
    penalty_default_kwargs = {"norm_matrix": norm_matrix}

    def get_invariant_matrix(self, rng, shape):
        if shape[0] != self.n_rows:
            raise ValueError("Shape must align with the norm matrix")
        return tl.ones(shape)

    def get_non_invariant_matrix(self, rng, shape):
        return tl.tensor(rng.random_sample(size=shape))

    def test_penalty(self, random_matrices):
        B01 = random_matrices[0][: self.n_rows // 2]
        B02 = random_matrices[0][self.n_rows // 2 :]
        penalty_manual = tl.sum((B01[1:] - B01[:-1]) ** 2) + tl.sum((B02[1:] - B02[:-1]) ** 2)

        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        assert penalty.penalty(random_matrices[0]) == pytest.approx(penalty_manual)

    def test_update_is_correct_on_example(self, rng):
        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        X = tl.tensor(rng.random_sample(size=(self.n_rows, 10)))
        feasibility_penalty = 5
        Y = penalty.factor_matrix_update(X, feasibility_penalty, None)

        aug_norm_matrix = self.norm_matrix + 0.5 * feasibility_penalty * tl.eye(self.n_rows)
        assert_allclose(Y, tl.solve(aug_norm_matrix, 2.5 * X), rtol=RTOL_SCALE * 1e-7)


@pytest.mark.skipif(
    tl.get_backend() != "numpy",
    reason="The TV penalty is only supported with the Numpy backend due to C dependencies",
)
class TestTotalVariationPenalty(BaseTestFactorMatrixPenalty):
    PenaltyType = penalties.TotalVariationPenalty
    penalty_default_kwargs = {"reg_strength": 1, "l1_strength": 0}
    min_rows = 3

    def get_invariant_matrix(self, rng, shape):
        return tl.zeros(shape)

    def get_non_invariant_matrix(self, rng, shape):
        return tl.tensor(rng.uniform(size=shape))

    def test_penalty(self, rng):
        shape = rng.randint(3, 20), rng.randint(3, 20)
        # shape = 10, 10
        tv_penalty = self.PenaltyType(reg_strength=1, l1_strength=1)

        # Penalty is 0 if input is 0
        assert tv_penalty.penalty(tl.zeros(shape)) == pytest.approx(0)

        # Penalty is sum(abs(X)) if X is only ones
        X1 = tl.ones(shape)
        assert tv_penalty.penalty(X1) == pytest.approx(tl.sum(tl.abs(X1)))

        # Penalty is sum(abs(X))+2 if X is only ones except for one non-boundary entry (not index 0 or -1) in one
        # column which is zero
        X2 = tl.ones(shape)
        X2[shape[0] // 2, shape[1] // 2] = 0
        print("X2", X2)
        assert tv_penalty.penalty(X2) == pytest.approx(tl.sum(tl.abs(X2)) + 2)

        # Penalty is sum(abs(X))+1 if X is only ones except for one boundary entry (index 0 or -1) in one column
        X3 = tl.ones(shape)
        X3[0, shape[1] // 2] = 0
        assert tv_penalty.penalty(X3) == pytest.approx(tl.sum(tl.abs(X3)) + 1)

        # Penalty is sum(abs(X))+n_cols if all columns of x is on form [0, 0, 0, 0, 1, 1, 1, 1]
        X4 = tl.ones(shape)
        X4[: shape[0] // 2] = 0
        assert tv_penalty.penalty(X4) == pytest.approx(tl.sum(tl.abs(X4)) + shape[1])

        # Penalty is zero for constant columns if l1_strength is 0
        tv_penalty_no_l1 = self.PenaltyType(reg_strength=1, l1_strength=0)
        X_constant_columns = rng.uniform(shape[0]) * tl.ones(shape)
        assert tv_penalty_no_l1.penalty(X_constant_columns) == pytest.approx(0)

    def test_input_is_checked(self):
        with pytest.raises(ValueError):
            tv_penalty = self.PenaltyType(reg_strength=0, l1_strength=1)
        with pytest.raises(ValueError):
            tv_penalty = self.PenaltyType(reg_strength=-1, l1_strength=1)
        with pytest.raises(ValueError):
            tv_penalty = self.PenaltyType(reg_strength=1, l1_strength=-1)

        tv_penalty = self.PenaltyType(reg_strength=1, l1_strength=0)

        HAS_TV = penalties.HAS_TV
        penalties.HAS_TV = False
        with pytest.raises(ModuleNotFoundError):
            tv_penalty = self.PenaltyType(reg_strength=1, l1_strength=0)  # pragma: noqa
        penalties.HAS_TV = HAS_TV

    def test_l1_is_applied(self):
        shape = (10, 3)
        normally_invariant_matrix = tl.ones(shape)
        penalty_without_l1 = self.PenaltyType(reg_strength=1, l1_strength=0)
        assert_allclose(
            normally_invariant_matrix, penalty_without_l1.factor_matrix_update(normally_invariant_matrix, 1, None)
        )

        penalty_with_l1 = self.PenaltyType(reg_strength=1, l1_strength=1000)
        assert_allclose(tl.zeros(shape), penalty_with_l1.factor_matrix_update(normally_invariant_matrix, 1, None))


class TestNonNegativity(MixinTestHardConstraint, BaseTestRowVectorPenalty):
    PenaltyType = penalties.NonNegativity

    def get_invariant_row(self, rng, n_columns):
        random_row = tl.tensor(rng.uniform(size=(1, n_columns)))
        return random_row

    def get_non_invariant_row(self, rng, n_columns):
        random_row = tl.tensor(rng.uniform(size=(1, n_columns)))
        random_row[0] = -1
        return random_row


@pytest.mark.skipif(
    tl.get_backend() != "numpy",
    reason=(
        "The unimodality constraint is only supported with the Numpy backend due"
        " to the serial nature of the unimodal regression algorithm and the implementation's use of Numba"
    ),
)
class TestUnimodality(MixinTestHardConstraint, BaseTestFactorMatrixPenalty):
    PenaltyType = penalties.Unimodality
    penalty_default_kwargs = {}
    min_rows = 3

    def get_invariant_matrix(self, rng, shape):
        matrix = tl.zeros(shape)
        I, J = shape
        t = np.linspace(-10, 10, I)
        for j in range(J):
            sigma = rng.uniform(0.5, 1)
            mu = rng.uniform(-5, 5)
            matrix[:, j] = stats.norm.pdf(t, loc=mu, scale=sigma)
        return matrix

    def get_non_invariant_matrix(self, rng, shape):
        # There are at least 3 rows
        M = rng.uniform(size=shape)
        M[1, :] = -1  # M is positive, so setting the second element to -1 makes it impossible for it to be unimodal
        return M

    @pytest.mark.parametrize("non_negativity", [True, False])
    def test_non_negativity_used(self, non_invariant_matrix, non_negativity):
        # Check that non_negativity is used
        with patch("matcouply.penalties.unimodal_regression") as mock:
            unimodality_constaint = self.PenaltyType(non_negativity=non_negativity)
            unimodality_constaint.factor_matrix_update(non_invariant_matrix, 1, None)
            mock.assert_called_once_with(non_invariant_matrix, non_negativity=non_negativity)


def test_unimodality_skipped():
    with patch("matcouply.decomposition.tensorly.get_backend", return_value="pytorch"):
        with pytest.raises(RuntimeError):
            penalties.Unimodality()

        with pytest.raises(RuntimeError):
            penalties.Unimodality(non_negativity=True)


class TestParafac2(BaseTestFactorMatricesPenalty):
    PenaltyType = penalties.Parafac2
    min_matrices = 2  # PARAFAC2 constraint does not make sense for only one matrix
    # The factor matrix needs at least as many rows as columns for PARAFAC2
    max_columns = 5
    min_rows = 5

    def test_projection_improves_with_num_iterations(self, random_rank5_ragged_cmf, rng):
        cmf, shapes, rank = random_rank5_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()
        feasibility_penalties = tl.tensor(rng.uniform(2, 3, size=len(B_is)))

        pf2_1it = self.PenaltyType(n_iter=1)
        pf2_5it = self.PenaltyType(n_iter=5)

        auxes_1it = pf2_1it.init_aux(matrices, rank, 1, rng)
        auxes_5it = ([tl.copy(Pi) for Pi in auxes_1it[0]], tl.copy(auxes_1it[1]))

        proj_1it = pf2_1it.factor_matrices_update(B_is, feasibility_penalties, auxes_1it)
        proj_5it = pf2_5it.factor_matrices_update(B_is, feasibility_penalties, auxes_5it)

        error_1it = sum(tl.sum(err**2) for err in pf2_1it.subtract_from_auxes(proj_1it, B_is))
        error_5it = sum(tl.sum(err**2) for err in pf2_5it.subtract_from_auxes(proj_5it, B_is))

        assert error_5it < error_1it

    def test_factor_matrices_update_invariant_point(self, rng):
        svd = get_svd("truncated_svd")

        # Construct invariant matrices in NumPy for double precision
        def random_orthogonal(size):
            X = rng.standard_normal(size)
            return np.linalg.qr(X)[0]

        deltaB = rng.standard_normal((3, 3))
        P_is = [random_orthogonal((10, 3)) for _ in range(5)]
        invariant_matrices = [P_i @ deltaB for P_i in P_is]

        deltaB = tl.tensor(deltaB)
        P_is = [tl.tensor(P_i) for P_i in P_is]
        invariant_matrices = [tl.tensor(B_i) for B_i in invariant_matrices]
        auxes = P_is, deltaB

        feasibility_penalties = [10] * len(invariant_matrices)
        pf2_penalty = penalties.Parafac2()

        out = pf2_penalty.factor_matrices_update(invariant_matrices, feasibility_penalties, auxes)
        assert_allclose(deltaB, out[1], rtol=1e-6 * RTOL_SCALE)
        for P_i, out_matrix in zip(P_is, out[0]):
            if tl.get_backend() == "numpy":
                rtol = 1e-6
            else:
                # This seems to be very unstable with single precision, one of the entries in one of the P_is is often too large
                rtol = 1e-2

            assert_allclose(P_i, out_matrix, rtol=rtol, err_msg="This can be somewhat unstable with single precision")

    def test_not_updating_basis_matrices_works(self, rng):
        svd = get_svd("truncated_svd")
        deltaB = tl.tensor(rng.standard_normal((3, 3)))
        P_is = [svd(tl.tensor(rng.standard_normal(size=(10, 3))), n_eigenvecs=3)[0] for _ in range(5)]
        wrong_P_is = [svd(tl.tensor(rng.standard_normal(size=(10, 3))), n_eigenvecs=3)[0] for _ in range(5)]
        B_is = [tl.matmul(P_i, deltaB) for P_i in P_is]
        auxes = wrong_P_is, deltaB

        feasibility_penalties = [10] * len(B_is)
        pf2_penalty = penalties.Parafac2(update_basis_matrices=False)

        out = pf2_penalty.factor_matrices_update(B_is, feasibility_penalties, auxes)
        assert not tl.all(deltaB == out[1])
        for P_i, out_matrix in zip(wrong_P_is, out[0]):
            assert_allclose(P_i, out_matrix)

    def test_not_updating_coordinate_matrix_works(self, rng):
        svd = get_svd("truncated_svd")
        deltaB = tl.tensor(rng.standard_normal((3, 3)))
        wrong_deltaB = tl.tensor(rng.standard_normal((3, 3)))
        P_is = [svd(tl.tensor(rng.standard_normal(size=(10, 3))), n_eigenvecs=3)[0] for _ in range(5)]
        B_is = [tl.matmul(P_i, deltaB) for P_i in P_is]
        auxes = P_is, wrong_deltaB

        feasibility_penalties = [10] * len(B_is)
        pf2_penalty = penalties.Parafac2(update_coordinate_matrix=False)
        out = pf2_penalty.factor_matrices_update(B_is, feasibility_penalties, auxes)
        assert_allclose(wrong_deltaB, out[1])
        for P_i, out_matrix in zip(P_is, out[0]):
            assert not tl.all(P_i == out_matrix)

    def test_factor_matrices_update_reduces_penalty(self, rng, random_matrices):
        svd = get_svd("truncated_svd")

        shapes = [tl.shape(matrix) for matrix in random_matrices]
        rank = shapes[0][1]
        deltaB = tl.tensor(rng.standard_normal((rank, rank)))
        P_is = [svd(tl.tensor(rng.standard_normal(size=shape)), n_eigenvecs=rank)[0] for shape in shapes]
        auxes = P_is, deltaB

        feasibility_penalties = [10] * len(random_matrices)
        pf2_penalty = penalties.Parafac2()

        initial_penalty = pf2_penalty.penalty(pf2_penalty.auxes_as_matrices(auxes))
        out = pf2_penalty.factor_matrices_update(random_matrices, feasibility_penalties, auxes)
        assert pf2_penalty.penalty(pf2_penalty.auxes_as_matrices(out)) <= initial_penalty

    def test_factor_matrices_update_changes_input(self, random_matrices, rng):
        svd = get_svd("truncated_svd")

        shapes = [tl.shape(matrix) for matrix in random_matrices]
        rank = shapes[0][1]
        deltaB = tl.tensor(rng.standard_normal((rank, rank)))
        P_is = [svd(tl.tensor(rng.standard_normal(size=shape)), n_eigenvecs=rank)[0] for shape in shapes]
        auxes = P_is, deltaB

        feasibility_penalties = [10] * len(random_matrices)
        pf2_penalty = penalties.Parafac2()

        out = pf2_penalty.factor_matrices_update(random_matrices, feasibility_penalties, auxes)
        constructed_out = [P_i @ out[1] for P_i in out[0]]
        for random_matrix, out_matrix in zip(random_matrices, constructed_out):
            assert not np.allclose(random_matrix, out_matrix)

    @pytest.mark.parametrize(
        "dual_init",
        ["random_uniform", "random_standard_normal", "zeros"],
    )
    @pytest.mark.parametrize("aux_init", ["random_uniform", "random_standard_normal", "zeros"])
    def test_rank_and_mode_validation_for_init_aux(self, rng, random_ragged_cmf, dual_init, aux_init):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()

        penalty = self.PenaltyType(aux_init=aux_init, dual_init=dual_init, **self.penalty_default_kwargs)
        # Test that mode and rank needs int input
        with pytest.raises(TypeError):
            penalty.init_aux(matrices, rank, mode=None)
        with pytest.raises(TypeError):
            penalty.init_aux(matrices, rank=None, mode=1)

        # Check that initialize fails with mode != 1
        with pytest.raises(ValueError):
            penalty.init_aux(matrices, rank, 0, random_state=rng)
        with pytest.raises(ValueError):
            penalty.init_aux(matrices, rank, 2, random_state=rng)

    @pytest.mark.parametrize("dual_init", ["random_uniform", "random_standard_normal", "zeros"])
    def test_uniform_init_aux(self, rng, random_ragged_cmf, dual_init):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()

        # Test that init works with random uniform init
        penalty = self.PenaltyType(aux_init="random_uniform", dual_init=dual_init, **self.penalty_default_kwargs)
        init_bases, init_coordinates = penalty.init_aux(matrices, rank, mode=1, random_state=rng)
        assert tl.shape(init_coordinates) == (rank, rank)
        assert tl.all(init_coordinates >= 0)
        assert tl.all(init_coordinates <= 1)

        for init_basis, B_i in zip(init_bases, B_is):
            assert_allclose(init_basis.T @ init_basis, tl.eye(rank))
            assert tl.shape(init_basis) == tl.shape(B_i)

    @pytest.mark.parametrize("dual_init", ["random_uniform", "random_standard_normal", "zeros"])
    def test_standard_normal_init_aux(self, rng, random_ragged_cmf, dual_init):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()

        # Test that init works with random standard normal init
        penalty = self.PenaltyType(
            aux_init="random_standard_normal", dual_init=dual_init, **self.penalty_default_kwargs
        )
        init_bases, init_coordinates = penalty.init_aux(matrices, rank, mode=1, random_state=rng)
        assert tl.shape(init_coordinates) == (rank, rank)
        for init_basis, B_i in zip(init_bases, B_is):
            assert_allclose(init_basis.T @ init_basis, tl.eye(rank))
            assert tl.shape(init_basis) == tl.shape(B_i)

    @pytest.mark.parametrize("dual_init", ["random_uniform", "random_standard_normal", "zeros"])
    def test_zeros_init_aux(self, rng, random_ragged_cmf, dual_init):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()

        # Test that init works with zeros init
        penalty = self.PenaltyType(aux_init="zeros", dual_init=dual_init, **self.penalty_default_kwargs)
        init_bases, init_coordinates = penalty.init_aux(matrices, rank, mode=1, random_state=rng)
        assert tl.shape(init_coordinates) == (rank, rank)
        assert_array_equal(init_coordinates, 0)
        for init_basis, B_i in zip(init_bases, B_is):
            assert_allclose(init_basis.T @ init_basis, tl.eye(rank), rtol=1e-6)
            assert tl.shape(init_basis) == tl.shape(B_i)

    @pytest.mark.parametrize("dual_init", ["random_uniform", "random_standard_normal", "zeros"])
    def test_given_init_aux(self, rng, random_ragged_cmf, dual_init):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()

        # Test that init works with specified init
        init_bases = [tl.eye(*tl.shape(B_i)) for B_i in B_is]
        init_coordinates = tl.tensor(rng.random_sample((rank, rank)))
        penalty = self.PenaltyType(
            aux_init=(init_bases, init_coordinates), dual_init=dual_init, **self.penalty_default_kwargs
        )
        init_bases_2, init_coordinates_2 = penalty.init_aux(matrices, rank, mode=1, random_state=rng)
        assert_array_equal(init_coordinates, init_coordinates_2)
        for init_basis, init_basis_2 in zip(init_bases, init_bases_2):
            assert_array_equal(init_basis, init_basis_2)

    @pytest.mark.parametrize("dual_init", ["random_uniform", "random_standard_normal", "zeros"])
    def test_validating_given_init_aux(self, rng, random_ragged_cmf, dual_init):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()

        # Test that init works with specified init
        init_bases = [tl.eye(*tl.shape(B_i)) for B_i in B_is]
        init_coordinates = tl.tensor(rng.random_sample((rank, rank)))
        # Test with various invalid basis matrix lists
        all_invalid_bases = []

        invalid_init_bases = copy(init_bases)
        invalid_init_bases[0] = tl.zeros(tl.shape(invalid_init_bases[0]))  # Not orthogonal is invalid
        all_invalid_bases.append(invalid_init_bases)

        invalid_init_bases = copy(init_bases)
        invalid_init_bases[0] = tl.zeros((tl.shape(invalid_init_bases[0])[0] + 1, rank))  # Wrong shape
        all_invalid_bases.append(invalid_init_bases)

        invalid_init_bases = copy(init_bases)
        invalid_init_bases[0] = tl.zeros((*tl.shape(invalid_init_bases[0]), 2))  # Wrong order
        all_invalid_bases.append(invalid_init_bases)

        all_invalid_bases.append(init_bases + init_bases)  # Wrong number of matrices

        for invalid_init_bases in all_invalid_bases:
            aux_init = (invalid_init_bases, init_coordinates)
            penalty = self.PenaltyType(aux_init=aux_init, dual_init=dual_init, **self.penalty_default_kwargs)

            with pytest.raises(ValueError):
                penalty.init_aux(matrices, rank, mode=1, random_state=rng)

        invalid_coordinates = [
            tl.zeros((rank, rank, rank)),
            tl.zeros((rank + 1, rank)),
            tl.zeros((rank, rank + 1)),
            tl.zeros((rank + 1, rank + 1)),
        ]
        for invalid_init_coordinates in invalid_coordinates:
            aux_init = (init_bases, invalid_init_coordinates)
            penalty = self.PenaltyType(aux_init=aux_init, dual_init=dual_init, **self.penalty_default_kwargs)

            with pytest.raises(ValueError):
                penalty.init_aux(matrices, rank, mode=1, random_state=rng)

    @pytest.mark.parametrize("dual_init", ["random_uniform", "random_standard_normal", "zeros"])
    def test_input_validation_for_init_aux(self, rng, random_ragged_cmf, dual_init):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()
        # Test that the init method must be a valid type
        invalid_inits = [
            None,
            1,
            1.1,
            (None, None),
            ([None] * len(matrices), tl.zeros((rank, rank))),
            ([tl.eye(J_i, rank) for J_i, k in shapes], None),
        ]
        for invalid_init in invalid_inits:
            penalty = self.PenaltyType(aux_init=invalid_init, dual_init=dual_init, **self.penalty_default_kwargs)
            with pytest.raises(TypeError):
                penalty.init_aux(matrices, rank, mode=1, random_state=rng)

        # Check that we get value error if aux init is str but not "random_uniform" or "random_standard_normal"
        penalty = self.PenaltyType(aux_init="invalid init name", dual_init=dual_init, **self.penalty_default_kwargs)
        for mode in range(2):
            with pytest.raises(ValueError):
                penalty.init_aux(matrices, rank, mode=mode, random_state=None)

    def test_subtract_from_aux(
        self,
    ):
        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        with pytest.raises(TypeError):
            penalty.subtract_from_aux(None, None)

    def test_subtract_from_auxes(self, rng, random_ragged_cmf):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()

        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        auxes = penalty.init_aux(matrices, rank, 1, random_state=rng)
        aux_matrices = penalty.auxes_as_matrices(auxes)

        aux_diffs = penalty.subtract_from_auxes(auxes, B_is)
        for i, B_i in enumerate(B_is):
            aux = aux_matrices[i]
            aux_diff = aux_diffs[i]

            assert_allclose(aux_diff, aux - B_i)

    def test_aux_as_matrix(self, random_matrix):
        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        # Check that this raises TypeError
        with pytest.raises(TypeError):
            penalty.aux_as_matrix(random_matrix)

    def test_auxes_as_matrices(self, rng, random_ragged_cmf):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        matrices = cmf.to_matrices()

        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        auxes = penalty.init_aux(matrices, rank, 1, random_state=rng)
        aux_matrices = penalty.auxes_as_matrices(auxes)

        for i, aux in enumerate(aux_matrices):
            aux = aux_matrices[i]
            basis_i = auxes[0][i]

            assert_allclose(aux, basis_i @ auxes[1])

    def test_penalty(self, random_ragged_cmf):
        cmf, shapes, rank = random_ragged_cmf
        weights, (A, B_is, C) = cmf
        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        assert penalty.penalty(B_is) == 0

        with pytest.raises(TypeError):
            penalty.penalty(A)


class TestTemporalSmoothness(BaseTestFactorMatricesPenalty):
    PenaltyType = penalties.TemporalSmoothnessPenalty
    penalty_default_kwargs = {"smoothness_l": 1}

    n_rows = 10
    min_rows = n_rows
    max_rows = n_rows

    def test_penalty(self, random_regular_cmf):
        # Check that the penalty term is computed correctly.

        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        cmf, random_ragged_shapes, rank = random_regular_cmf
        weights, (A, B_is, C) = cmf

        K = len(B_is)

        penalty_term = 0
        for k in range(1, K):
            penalty_term += tl.sum((B_is[k] - B_is[k - 1]) ** 2)

        assert_allclose(penalty.penalty(B_is), penalty_term)

    def get_invariant_matrices(self, rng, shapes):
        # Generate a list of invariant matrices that will not be changed by the proximal operator.

        K = len(shapes)

        invariant_matrix = np.array(rng.random_sample(shapes[0]))
        invariant_matrices = [tl.tensor(invariant_matrix.copy()) for _ in range(K)]
        return invariant_matrices

    def get_non_invariant_matrices(self, rng, shapes):
        # Generate a list of invariant matrices that will be changed by the proximal operator.

        K = len(shapes)

        non_invariant_matrices = [tl.tensor(rng.random_sample(shapes[k])) for k in range(K)]
        return non_invariant_matrices

    def test_A_assembly(self, random_regular_cmf, rng):
        # Ensure A is assembled correctly by comparing the efficent method (A1) with the manual assembly (A2).

        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        cmf, random_ragged_shapes, rank = random_regular_cmf
        weights, (A, B_is, C) = cmf

        I = len(B_is)

        rhos = rng.random_sample(I)

        A1 = (
            tl.diag(tl.tensor([penalty._get_laplace_coef(i, I) + rho for i, rho in enumerate(rhos)]), k=0)
            - tl.diag(tl.ones(I - 1) * 2 * penalty.smoothness_l, k=1)
            - tl.diag(tl.ones(I - 1) * 2 * penalty.smoothness_l, k=-1)
        )

        A2 = tl.zeros((len(B_is), len(B_is)))

        for i in range(len(B_is)):
            for j in range(len(B_is)):
                if i == j:
                    A2[i, j] = 4 * penalty.smoothness_l + rhos[i]
                elif (i == j - 1) or (i == j + 1):
                    A2[i, j] = -2 * penalty.smoothness_l
                else:
                    pass

        A2[0, 0] -= 2 * penalty.smoothness_l
        A2[len(B_is) - 1, len(B_is) - 1] -= 2 * penalty.smoothness_l

        if tl.get_backend() == "numpy":
            assert_allclose(A1, A2)
        else:  # pytorch is slightly less accurate here, so we relax the tolerance
            assert_allclose(A1, A2, rtol=1e-5)


class TestLDSPenalty(BaseTestFactorMatricesPenalty):

    PenaltyType = penalties.LDSPenalty
    randomH = tl.random.random_tensor(shape=(10, 10))
    penalty_default_kwargs = {"smoothness_l": 1, "H": randomH}

    n_rows = 10
    min_rows = n_rows
    max_rows = n_rows

    def test_penalty(self, rng, random_regular_cmf):

        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        cmf, random_ragged_shapes, rank = random_regular_cmf
        weights, (A, B_is, C) = cmf

        K = len(B_is)
        J = B_is[0].shape[0]
        penalty.H = penalty.H[:J, :J]  # Ensure H is of the correct shape

        penalty_term = 0
        for k in range(1, K):
            penalty_term += tl.sum((B_is[k] - penalty.H @ B_is[k - 1]) ** 2)

        assert_allclose(penalty.penalty(B_is), penalty_term)

    def get_invariant_matrices(self, rng, shapes):
        """Generate a list of invariant matrices that will not be changed by the proximal operator."""

        penalty = self.PenaltyType(**self.penalty_default_kwargs)
        K = len(shapes)

        invariant_matrices = [np.array(rng.random_sample(shapes[0]))]

        for K in range(1, K):
            invariant_matrices.append(penalty.H @ invariant_matrices[-1])

        return invariant_matrices

    def get_non_invariant_matrices(self, rng, shapes):
        """Generate a list of invariant matrices that will be changed by the proximal operator."""

        K = len(shapes)

        non_invariant_matrices = [tl.tensor(rng.random_sample(shapes[k])) for k in range(K)]
        return non_invariant_matrices



# -*- coding: utf-8 -*-
# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

__author__ = """Marie Roald"""
__email__ = "roald.marie@gmail.com"
__version__ = "0.2.0"


from . import coupled_matrices, data, decomposition, penalties, random



# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

import numpy as np

try:
    from numba import jit
except ImportError:  # pragma: no cover

    def jit(*args, **kwargs):
        return lambda x: x


@jit(nopython=True, cache=True)
def _merge_intervals_inplace(merge_target, merger, sum_weighted_y, sum_weighted_y_sq, sum_weights, level_set):
    sum_weighted_y[merge_target] += sum_weighted_y[merger]
    sum_weighted_y_sq[merge_target] += sum_weighted_y_sq[merger]
    sum_weights[merge_target] += sum_weights[merger]

    # Update the level set
    level_set[merge_target] = sum_weighted_y[merge_target] / sum_weights[merge_target]


@jit(
    nopython=True, cache=True,
)
def prefix_isotonic_regression(y, weights=None, non_negativity=False):
    if weights is None:
        weights = np.ones_like(y)

    sumwy = weights * y
    sumwy2 = weights * y * y
    sumw = weights.copy()

    level_set = np.zeros_like(y)
    index_range = np.zeros_like(y, dtype=np.int32)
    error = np.zeros(y.shape[0] + 1)  # +1 since error[0] is error of empty set

    level_set[0] = y[0]
    index_range[0] = 0
    num_samples = y.shape[0]

    if non_negativity:
        cumsumwy2 = np.cumsum(sumwy2)
        threshold = np.zeros(level_set.shape)
        if level_set[0] < 0:
            threshold[0] = True
            error[1] = cumsumwy2[0]

    for i in range(1, num_samples):
        level_set[i] = y[i]
        index_range[i] = i
        while level_set[i] <= level_set[index_range[i] - 1] and index_range[i] != 0:
            _merge_intervals_inplace(i, index_range[i] - 1, sumwy, sumwy2, sumw, level_set)
            index_range[i] = index_range[index_range[i] - 1]

        levelerror = sumwy2[i] - (sumwy[i] ** 2 / sumw[i])
        if non_negativity and level_set[i] < 0:
            threshold[i] = True
            error[i + 1] = cumsumwy2[i]
        else:
            error[i + 1] = levelerror + error[index_range[i]]

    if non_negativity:
        for i in range(len(level_set)):
            if threshold[i]:
                level_set[i] = 0

    return (level_set, index_range), error


@jit(nopython=True, cache=True)
def _compute_isotonic_from_index(end_index, level_set, index_range):
    idx = end_index - 1
    y_iso = np.empty_like(level_set[: idx + 1]) * np.nan

    while idx >= 0:
        y_iso[index_range[idx] : idx + 1] = level_set[idx]
        idx = index_range[idx] - 1

    return y_iso


def _get_best_unimodality_index(error_left, error_right):
    best_error = error_right[-1]
    best_idx = 0
    for i in range(error_left.shape[0]):
        error = error_left[i] + error_right[len(error_left) - i - 1]
        if error < best_error:
            best_error = error
            best_idx = i
    return best_idx, best_error


def _unimodal_regression(y, non_negativity):
    iso_left, error_left = prefix_isotonic_regression(y, non_negativity=non_negativity)
    iso_right, error_right = prefix_isotonic_regression(y[::-1], non_negativity=non_negativity)

    num_samples = y.shape[0]
    best_idx, error = _get_best_unimodality_index(error_left, error_right)
    y_iso_left = _compute_isotonic_from_index(best_idx, iso_left[0], iso_left[1])
    y_iso_right = _compute_isotonic_from_index(num_samples - best_idx, iso_right[0], iso_right[1])

    return np.concatenate([y_iso_left, y_iso_right[::-1]]), error


def unimodal_regression(y, non_negativity=False):
    r"""Compute the unimodal vector, :math:`\mathbf{u}` that minimizes :math:`\|\mathbf{y} - \mathbf{u}\|`.

    The unimodal regression problem is a problem on the form

    .. math:: \min_{\mathbf{u}} \|\mathbf{y} - \mathbf{u}\|\\
              \text{s.t.} u_1 \leq u_2 \leq ... \leq u_{t-1} \leq u_t \geq u_{t+1} \geq ... \geq u_{n-1} \geq u_n,

    for some index :math:`1 \leq t \leq n`. That is, it projects the input vector :math:`\mathbf{y}` onto the
    set of unimodal vectors. The *unimodal regression via prefix isotonic regression*  algorithm :cite:p:`stout2008unimodal`
    is used to efficiently solve the unimodal regression problem.

    Parameters
    ----------
    y : ndarray
        Vector to project. If it is an N-dimensional array, then it projects the first-mode fibers
        (e.g. columns in the case of matrices).
    non_negativity : bool
        If True, then non-negativity is imposed

    Returns
    -------
    ndarray
        The unimodal vector or array of unimodal vectors
    """
    y = np.asarray(y)
    if y.ndim == 1:
        return _unimodal_regression(y, non_negativity=non_negativity)[0]
    else:
        y2 = np.ascontiguousarray(y)
        y2 = y2.reshape(y2.shape[0], -1)
        unfolded_output = np.stack(
            [_unimodal_regression(y2[:, r], non_negativity=non_negativity)[0] for r in range(y2.shape[1])], axis=1,
        )
        return unfolded_output.reshape(y.shape)



# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

import tensorly as tl

from .coupled_matrices import CoupledMatrixFactorization


def random_coupled_matrices(
    shapes, rank, full=False, random_state=None, normalise_factors=True, normalise_B=False, **context
):
    """Generate a random coupled matrix decomposition (with non-negative entries)

    Parameters
    ----------
    shapes : tuple
        A tuple where each element represents the shape of a matrix
        represented by the coupled matrix decomposition model. The
        second element in each shape-tuple must be constant.
    rank : int or int list
        rank of the coupled matrix decomposition
    full : bool, optional, default is False
        if True, a list of dense matrices is returned otherwise,
        the decomposition is returned
    random_state : `np.random.RandomState`

    Examples
    --------
    Here is an example of how to generate a random coupled matrix factorization

    >>> from matcouply.random import random_coupled_matrices
    >>> shapes = ((5, 10), (6, 10), (7, 10))
    >>> cmf = random_coupled_matrices(shapes, rank=4)
    >>> print(cmf)
    (weights, factors) : rank-4 CoupledMatrixFactorization of shape ((5, 10), (6, 10), (7, 10))
    """
    rns = tl.check_random_state(random_state)
    if not all(shape[1] == shapes[0][1] for shape in shapes):
        raise ValueError("All matrices must have equal number of columns.")

    A = tl.tensor(rns.random_sample((len(shapes), rank), **context))
    B_is = [tl.tensor(rns.random_sample((j_i, rank), **context)) for j_i, k in shapes]
    K = shapes[0][1]
    C = tl.tensor(rns.random_sample((K, rank), **context))

    weights = tl.ones(rank, **context)

    if normalise_factors or normalise_B:
        B_i_norms = [tl.norm(B_i, axis=0) for B_i in B_is]
        B_is = [B_i / B_i_norm for B_i, B_i_norm in zip(B_is, B_i_norms)]
        B_i_norms = tl.stack(B_i_norms)

        A = A * B_i_norms
    if normalise_factors:
        A_norm = tl.norm(A, axis=0)
        A = A / A_norm

        C_norm = tl.norm(C, axis=0)
        C = C / C_norm
        weights = A_norm * C_norm

    cmf = CoupledMatrixFactorization((weights, (A, B_is, C)))
    if full:
        return cmf.to_matrices()
    else:
        return cmf
